/*
 * employer.c
 *
 * part of dipcd source code
 *
 * Copyright (C) Kamran Karimi
 */

#include "dipcd.h"
#include "funcs.h"

/* employer is forked whenever some request has to be satisfied remotely.
   It connects to the remote machine (sometimes this 'remote' machine can 
   be the local computer, like when a process on this computer wants to 
   refere to the manager of a distributed shared memory that is running in
   the same machine), informs it of the request and then waits for a reply.
   It uses Internet sockets to send the request, but relies on UNIX domain 
   sockets for the reply (through the front_end) The results are delivered 
   to the original process that caused the creation of the employer */ 

extern struct sockaddr_in MY_ADDRESS, REFEREE_ADDRESS;
extern int verbose;
extern int global_timeout;
extern int employer_worker_timeout;
extern int employer_shm_timeout;
extern int employer_referee_timeout;

static int orig_pid = 0;

static char ipc_type[3][4] =
{
 "SHM",
 "MSG",
 "SEM"
};

static int ux_sockfd;
static char ux_name[80];
static struct dipc_info *global_di;

/* return the results to the kernel */
static void return_result(struct dipc_info *di, char *data)
{
 void *pt[2];

 pt[0] = di;
 pt[1] = data;
 if(ux_name[0])
  unlink(ux_name);
 if(orig_pid != 0)
  di->pid = orig_pid; 
 if(dipc(di->pid, DIPC_RETWORK, (int *)pt) < DIPC_SUCCESS)
 {
  sprintf(mstr,"ERROR: DIPC_RETWORK failed");
  mess(mstr, INFO);
 }
}
 
static void sig_handler(int signo)
{
 if(verbose)
 {
  sprintf(mstr,"Quiting For Signal No. %d",signo); 
  mess(mstr,INFO);
 } 
 /* do last works */
 global_di->arg1 = -EINTR;
 return_result(global_di, NULL);
 exit(5);
}

static void exit_handler(int Code, void *arg)
{
 sprintf(mstr,"Exited With Code %d",Code); 
 mess(mstr,INFO);
}

 
void employer(struct dipc_info *di, char *data)
{
 fd_set rSelFD; /* for select */
 struct timeval TimeVal;
 struct message message;
 struct sockaddr_in owner_addr;
 struct sockaddr_un ux_addr, ux_client_addr;
 int ux_len, ux_client_len;
 int in_sock, ux_newsockfd;
 struct sigaction sact;
 int want_trans, act_trans;
 struct passwd *pw;
 struct usr_info uinfo;
 int sel_res, TimeOut;
 
 if(verbose)
 {
  sprintf(mstr,"Started with Function %s type %s",
             work_string(di->func), 
             (di->type < 3) ? ipc_type[di->type] : "N/A");
  mess(mstr, INFO);
 }
 global_di = di; /* for interrupt handler */
 sact.sa_handler = sig_handler;
 sigemptyset(&sact.sa_mask);
 sact.sa_flags = 0;
 
 if(on_exit(exit_handler, NULL) < 0)
 {
  sprintf(mstr,"ERROR: Can't install exit handler BECAUSE %s",strerror(errno)); 
  mess(mstr, INFO);
 }
 ux_name[0] = 0;
 
 if(sigaction(SIGINT, &sact, NULL) < 0)
 {
  sprintf(mstr,"WARNING: Can't Catch Signal  BECAUSE  %s",strerror(errno));
  mess(mstr, INFO);
 } 
 if(sigaction(SIGTERM, &sact, NULL) < 0)
 {
  sprintf(mstr,"WARNING: Can't Catch Signal  BECAUSE  %s",strerror(errno));
  mess(mstr, INFO);
 } 
 if(sigaction(SIGSEGV, &sact, NULL) < 0)
 {
  sprintf(mstr,"ERROR: Can't Catch Signal BECAUSE %s",strerror(errno));
  mess(mstr,INFO);
 } 
 /* first set up a UNIX domain socket to receive the results */
 if(di->type < 3)
  sprintf(ux_name,"%sDIPC_%s.%d.%d",
             UNIX_PATH,ipc_type[di->type],di->arg1,getpid());
 else
 {
  di->arg1 = -DIPC_E_TYPE;
  sprintf(mstr,"ERROR: Unknown Type");
  mess(mstr, INFO);
  return_result(di, NULL);
  exit(20);
 } 
 bzero((char *)&ux_addr,sizeof(ux_addr));
 ux_addr.sun_family = AF_UNIX;
 strcpy(ux_addr.sun_path,ux_name);
 ux_len = strlen(ux_addr.sun_path) + sizeof(ux_addr.sun_family); 
 unlink(ux_name);                             
 ux_sockfd = plug(AF_UNIX, (struct sockaddr *) &ux_addr, ux_len);
 if(ux_sockfd < 0)
 {
  di->arg1 = -errno; 
  sprintf(mstr,"ERROR: Can't Create Socket %s",ux_name);
  mess(mstr, INFO);
  return_result(di, NULL);
  exit(20);
 } 
 /* prepare the dipcd message to be sent */      
 orig_pid = di->pid;
 message.info = *di;
 switch(message.info.func)
 {
  case WANTRD:
  case WANTWR:
   message.request = REQ_SHMMAN;
   TimeOut = employer_shm_timeout;
   break;
   
  case RMID:
  case DELKEY:
  case SEARCH:
  case COMMIT:
   message.request = REQ_REFEREE;
   TimeOut = employer_referee_timeout;
   break;
   
  case SEMCTL:
  case SEMOP:
  case MSGCTL:
  case MSGSND:
  case MSGRCV:
  case SHMCTL:
  case TSTGET: 
   message.request = REQ_DOWORK;
   TimeOut = employer_worker_timeout;
   break;
  
  default:
   di->arg1 = -errno; 
   sprintf(mstr,"ERROR: Unknown funcion");
   mess(mstr, INFO);
   return_result(di, NULL);
   exit(20); 
 }
 message.arch = MY_DIPC_ARCH;
 message.address = MY_ADDRESS;
 message.pid = getpid();
 
 if(message.request == REQ_REFEREE) 
  message.info.owner = REFEREE_ADDRESS.sin_addr.s_addr;
  
 bzero((char *) &owner_addr, sizeof(struct sockaddr_in));
 owner_addr.sin_family = AF_INET;
 owner_addr.sin_addr.s_addr = message.info.owner;
 if(message.request == REQ_REFEREE)
  owner_addr.sin_port = htons(DIPC_REFEREE_PORT);
 else 
  owner_addr.sin_port = htons(DIPC_TCP_PORT);
 if(verbose)
 {
  sprintf(mstr,"Destination Address Is: %s", inet_ntoa(owner_addr.sin_addr));
  mess(mstr, INFO);
 }
 
 /* get user name */
 if(message.info.uid != 0)
 {
  pw = getpwuid(message.info.uid);
  if(pw)
   strcpy(uinfo.name, pw->pw_name);
 } 
 if((message.info.uid == 0) || (pw == NULL)) /* no super user! */
  strcpy(uinfo.name, "dipcd");
 
 in_sock = ring(AF_INET, (struct sockaddr *)&owner_addr,
                                                      sizeof(owner_addr));
 if(in_sock < 0) 
 {
  message.info.arg1 = -errno;
  sprintf(mstr,"ERROR: get_in() Failed  BECAUSE  %s",strerror(errno));
  mess(mstr, INFO);
  return_result(&(message.info), NULL);
  exit(20);
 }
 want_trans = sizeof(message);
 act_trans = writen(in_sock, (char *)&message, want_trans);
 if(act_trans < want_trans)
 {
  message.info.arg1 = -errno;
  sprintf(mstr,"ERROR: Sent %d From %d Bytes  BECAUSE  %s",
          act_trans, want_trans,strerror(errno));
  mess(mstr, INFO);        
  return_result(&(message.info), NULL);
  exit(20);
 } 
 if(message.request == REQ_DOWORK)
 {
  /* send the user name */
  want_trans = sizeof(struct usr_info);
  act_trans = writen(in_sock, (char *)&uinfo, want_trans);
  if(act_trans < want_trans)
  {
   message.info.arg1 = -errno;
   sprintf(mstr,"ERROR: Sent %d From %d Bytes  BECAUSE  %s",
           act_trans, want_trans,strerror(errno));
   mess(mstr, INFO);        
   return_result(&(message.info), NULL);
   exit(20);
  }
 }
 /* send data for msgsnd() and  msgctl() with IPC_SET */ 
 if((message.info.func == MSGSND) ||
    ((message.info.func == MSGCTL) && (message.info.arg2 == IPC_SET)) ||
    ((message.info.func == SHMCTL) && (message.info.arg2 == IPC_SET)) ||
    (message.info.func == SEMOP) || 
    ((message.info.func == SEMCTL) && (message.info.arg3 == IPC_SET)) ||
    ((message.info.func == SEMCTL) && (message.info.arg3 == SETALL)))
 {
  want_trans = message.info.dsize;   
  act_trans = writen(in_sock, data, want_trans);
  if(act_trans < want_trans)
  {
   message.info.arg1 = -errno;
   sprintf(mstr,"ERROR: Sent %d From %d Bytes  BECAUSE %s",
           act_trans, want_trans, strerror(errno));
   mess(mstr, INFO);        
   return_result(&(message.info), NULL);
   exit(20);
  }
 }  
 close(in_sock); 
 if(verbose)
 {
  sprintf(mstr,"Request Sent");
  mess(mstr,INFO);
 }
 ux_client_len = sizeof(ux_client_addr);
 
 TimeVal.tv_sec = TimeOut;
 TimeVal.tv_usec = 0;
 FD_ZERO(&rSelFD);
 FD_SET(ux_sockfd,&rSelFD);
 if(TimeOut <= 0 || global_timeout == 0)
  sel_res = select(FD_SETSIZE, &rSelFD, NULL, NULL, NULL);
 else
  sel_res = select(FD_SETSIZE, &rSelFD, NULL, NULL, &TimeVal); 
 if( sel_res < 0)
 {
  message.info.arg1 = -errno;
  sprintf(mstr,"ERROR: select() Failed  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
  return_result(&(message.info), NULL);
  exit(20);
 }  
 if(FD_ISSET(ux_sockfd,&rSelFD))
 {
  if((ux_newsockfd = accept(ux_sockfd, (struct sockaddr *)&ux_client_addr,
                                                       &ux_client_len)) < 0)
  {
   message.info.arg1 = -errno;
   sprintf(mstr,"ERROR: accept() Failed  BECAUSE %s",strerror(errno));
   mess(mstr, INFO);
   return_result(&(message.info), NULL);
   exit(20);
  }
 
  /* Now message will be over written */
  want_trans = sizeof(message);
  act_trans = readn(ux_newsockfd, (char *)&message, want_trans);
  if(act_trans < want_trans)
  {
   message.info.arg1 = -errno;
   sprintf(mstr,"ERROR: Received %d From %d Bytes BECAUSE %s",
           act_trans, want_trans,strerror(errno));
   mess(mstr, INFO);        
   return_result(&(message.info), NULL);
   exit(20);
  } 
  /* receive data for msgrcv() and  msgctl() with IPC_STAT */ 
  if((message.info.func == MSGRCV) ||
     ((message.info.func == MSGCTL) && (message.info.arg2 == IPC_STAT)) ||
     ((message.info.func == SHMCTL) && (message.info.arg2 == IPC_STAT)) ||
     
     ((message.info.func == SEMCTL) && (message.info.arg3 == IPC_STAT)) ||
     ((message.info.func == SEMCTL) && (message.info.arg3 == GETALL)))
  {
   want_trans = message.info.dsize;   
   act_trans = readn(ux_newsockfd, data, want_trans);
   if(act_trans < want_trans)
   {
    message.info.arg1 = -errno;
    sprintf(mstr,"ERROR: Received %d From %d Bytes  BECAUSE %s",
            act_trans, want_trans,strerror(errno));
    mess(mstr,INFO);
    return_result(&(message.info), NULL);
    exit(20);
   } 
  }   
  close(ux_newsockfd);
  close(ux_sockfd);
  if(verbose)
  {
   sprintf(mstr,"Reply Received. Result is %d",message.info.arg1);
   mess(mstr,INFO);
  }
   
  if(message.info.func == COMMIT && (message.info.arg3 & IPC_OWN) && 
       message.info.type == SHM)
  {
   unlink(ux_name);
   shm_man(&(message.info), data);
   exit(0);
  }   
  return_result(&(message.info), data);
 }
 else
 { /* time out */
  sprintf(mstr,"ERROR: No Reply");
  mess(mstr,INFO);
  message.info.arg1 = -DIPC_E_TIMEOUT;
  return_result(&(message.info), data);
  exit(20);
 }   
 exit(0);
}
