/*
 * front_end.c
 *
 * part of dipcd source code
 *
 * Copyright (C) Kamran Karimi
 */

#include "dipcd.h"
#include "funcs.h"

/* front_end is the door keeper of dipcd. All incoming messages over the 
   network pass through it. The incoming messages can either be results 
   for requests made from dipcd processes in this computer, or they can 
   be requests from processes on other machines for work in this computer.
   In the first case, the front_end passes the messages to its proper
   process. in the second case a new worker processes is forked. The new 
   worker inherits the front_end's socket file descriptor, and so can 
   exchange information directly with the remote process */

extern int verbose;

static int in_sockfd, in_newsockfd;

static char ipc_type[3][4] =
{
 "SHM",
 "MSG",
 "SEM"
};

/* the newly arrived dipcd message is for a process that is waiting for it.
   Use a UNIX domain socket to deliver it. */
static BOOLEAN pass_message(int in_sock, struct message *message)
{
 int want_trans, act_trans;
 char temp[40];
 char ux_name[80];
 int ux_sockfd, ux_len;
 struct sockaddr_un ux_addr; 
 char data[MAXDATA];
 
 if(message->info.type < 3)
  sprintf(temp,"%sDIPC_%s",UNIX_PATH, ipc_type[message->info.type]);
 else
 { 
  sprintf(mstr,"ERROR: Unknown Type");
  mess(mstr, INFO);
  return FAIL;
 }
 /* other processes don't know the shared memory manager's */
 if(message->request == REQ_SHMMAN)  
  sprintf(ux_name,"%s.%d",temp,message->info.key);
 else
  sprintf(ux_name,"%s.%d.%d",temp,message->info.key, message->pid);
 if(verbose)
 { 
  sprintf(mstr,"Result Came for socket %s",ux_name);
  mess(mstr, INFO);
 }

 bzero((char *) &ux_addr,sizeof(ux_addr));
 ux_addr.sun_family = AF_UNIX;
 strcpy(ux_addr.sun_path,ux_name);
 ux_len = strlen(ux_addr.sun_path) + sizeof(ux_addr.sun_family);                                
 ux_sockfd = ring(AF_UNIX, (struct sockaddr *) &ux_addr, ux_len);
 if(ux_sockfd < 0)
 {
  sprintf(mstr,"ERROR: Can't Connect To Waiting Task");
  mess(mstr, INFO);
  return FAIL;
 } 
 want_trans = sizeof(*message);
 act_trans = writen(ux_sockfd,(char *)message, want_trans);
 if(act_trans < want_trans)
 {
  sprintf(mstr,"ERROR: Sent %d From %d Bytes  BECAUSE %s",
          act_trans,want_trans,strerror(errno));
  mess(mstr, INFO);        
  return FAIL;
 } 
 /* see if there is any data other than the message */ 
 if((message->info.func == MSGRCV) ||
    ((message->info.func == MSGCTL) && (message->info.arg2 == IPC_STAT)) ||
    ((message->info.func == SHMCTL) && (message->info.arg2 == IPC_STAT)))
 {
  want_trans = message->info.dsize;
  act_trans = get_in_data(in_sock, data, message);
  if(act_trans < want_trans)
  {
   sprintf(mstr,"ERROR: Received %d From %d Bytes  BECAUSE %s",
           act_trans, want_trans,strerror(errno));
   mess(mstr, INFO);        
   return FAIL;
  }     
  want_trans = message->info.dsize;
  act_trans = writen(ux_sockfd, data, message->info.dsize);
  if(act_trans < want_trans)
  {
   sprintf(mstr,"ERROR: Sent %d From %d Bytes  BECAUSE %s",
           act_trans, want_trans,strerror(errno));
   mess(mstr, INFO);
   return FAIL;
  }
 }
 close(ux_sockfd);
 return SUCCESS;
}


static void sig_handler(int signo)
{
 if(verbose)
 {
  sprintf(mstr,"Quiting For Signal No. %d",signo);
  mess(mstr, INFO);
 } 
 /* do last works */ 
 exit(5);
}

void front_end(void)
{
 fd_set rSelFD;
 struct timeval TimeVal;
 int want_trans, act_trans;
 int in_client_len, in_len;
 struct sockaddr_in in_addr, in_client_addr;
 struct message message;
 struct sigaction sact;
 int child_pid;
 
 if(verbose)
 {
  sprintf(mstr,"front_end Started");
  mess(mstr, INFO);
 }
 sact.sa_handler = sig_handler;
 sigemptyset(&sact.sa_mask);
 sact.sa_flags = 0;
 
 if(sigaction(SIGINT, &sact, NULL) < 0)
 {
  sprintf(mstr,"ERROR: Can't Catch Signal  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
 } 
 if(sigaction(SIGTERM, &sact, NULL) < 0)
 {
  sprintf(mstr,"ERROR: Can't Catch Signal  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
 }
 if(sigaction(SIGSEGV, &sact, NULL) < 0)
 {
  sprintf(mstr,"ERROR: Can't Catch Signal BECAUSE %s",strerror(errno));
  mess(mstr,INFO);
 }
 sact.sa_handler = child_handler;
 if(sigaction(SIGCHLD, &sact, NULL) < 0)
 {
  sprintf(mstr,"ERROR: Can't Catch Signal  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
 } 
  
 bzero((char *)&in_addr,sizeof(in_addr));
 in_addr.sin_family = AF_INET;
 in_addr.sin_addr.s_addr = htonl(INADDR_ANY);
 in_addr.sin_port = htons(DIPC_TCP_PORT);
 in_len = sizeof(in_addr);

 in_sockfd = plug(AF_INET, (struct sockaddr *) &in_addr, in_len);
 if(in_sockfd < 0)
 {
  sprintf(mstr,"ERROR: Can't Create Socket");
  mess(mstr, INFO);
  exit(20);
 }

 /* check interrupts from now on */
 while(1) 
 {
  TimeVal.tv_sec = 0;
  TimeVal.tv_usec = 500000; /* this time out is unused for now */
  FD_ZERO(&rSelFD);
  FD_SET(in_sockfd,&rSelFD);
  while(1)
  {
   if(select(FD_SETSIZE,&rSelFD,NULL,NULL,&TimeVal) == -1)
   {
    if(errno != EINTR)
    {
     sprintf(mstr,"ERROR: select() Failed  BECAUSE %s",strerror(errno));
     mess(mstr, INFO);
     close(in_sockfd);
     exit(20);
    }
   }
   else break;
  }

  if(FD_ISSET(in_sockfd,&rSelFD))
  { 
   in_client_len = sizeof(in_client_addr);
   while(1)
   {
    if((in_newsockfd = accept(in_sockfd, (struct sockaddr *)&in_client_addr,
                                                      &in_client_len)) < 0)
    {
     if(errno != EINTR)
     {
      sprintf(mstr,"ERROR: accept() Failed  BECAUSE %s",strerror(errno));
      mess(mstr, INFO);
      close(in_sockfd);
      exit(20);
     } 
    }
    else break;
   }
   want_trans = sizeof(message);
   act_trans = get_in_message(in_newsockfd,&message);
   if(act_trans < want_trans)
   {
    sprintf(mstr,"ERROR: Received %d From %d Bytes  BECAUSE %s",
            act_trans, want_trans,strerror(errno));
    mess(mstr, INFO); 
    exit(20);
   }
   if(verbose)
   {
    sprintf(mstr,"Remote Call From %s",inet_ntoa(message.address.sin_addr)); 
    mess(mstr, INFO);
   }
   switch(message.request)
   {
    case RES_REFEREE:
    case RES_DOWORK:
    case RES_SHMMAN:
    case REQ_SHMMAN: /* for shm_man */
     pass_message(in_newsockfd, &message);
     break;
    default: 
     if((child_pid = fork()) < 0)
     {
      sprintf(mstr,"ERROR: fork() failed  BECAUSE %s",strerror(errno));
      mess(mstr, INFO);
     }  
     else if(child_pid == 0)
     {
      /* give the worker the file descriptor */
      worker(in_newsockfd, message);
      exit(20);
     }
     break; 
   } 
   close(in_newsockfd); 
  }
  /* else ;  time out */
 }
 exit(0);
}
