/*
 * shm_man.c
 *
 * part of dipcd source code
 *
 * Copyright (C) Kamran Karimi
 */
	
#include "dipcd.h"
#include "funcs.h"
#include "back_door.h"

char back_ux_name[80];

/* shm_man is called by the employer that serviced the successful shmget() 
   system call. When an employer find out that it has helped make a machine 
   the owner of a shared memory, it will call the shm_man, to manage this 
   memory segment. */

/* for reader and writer queues */
#define SHM_READ  0
#define SHM_WRITE 1

/* There are two separete queues for readers and writers. When a machine 
   sends a request to the shared memory manager, it is put in the 
   appropriate WANT_ACCESS queue. After some time (or may be immediately) 
   the manager will try to satisfy the request. It may revoke other 
   machines' access to the shared memory and send the shared memory 
   contents to the requester. The manager then places the machine in the 
   PENDING queue and waits for an aknowledgment from it, to make sure it 
   has started using the shared memory. If the request does not arrive 
   within the shm_ak_timeout, it is assumed that the machine has some 
   problems, and its request is discarded. Otherwise it is turned into an
   ACCESSOR */

#define WANT_ACCESS 0
#define PENDING     1
#define ACCESSOR    2

extern struct sockaddr_in MY_ADDRESS, REFEREE_ADDRESS;
extern int verbose;
extern int global_timeout;
extern int shm_ak_timeout;
extern int shm_hold_timeout;

static clock_t last_change; /* tick value at last shm change */
static double ticks_sec; /* ticks per second */

static int ux_sockfd;
static char ux_name[80];

struct accessor {
 struct sockaddr_in address;
 int pid;
 time_t time;
 struct accessor *next,*prev;
};

/* six queues: 3 for WANT_ACCESS, PENDING and ACCESSOR readers,
  and 3 similar queues for writers */
struct accessor accessors_first[3][2];
struct accessor *accessors_last[3][2];

/* informatin about the managed shared memory  */ 
struct {
 key_t key;
 int id;
 char *attach;
 int size;
} my_shm;


static void sig_handler(int signo)
{
 if(verbose)
 {
  sprintf(mstr,"Quiting By Signal No. %d",signo);
  mess(mstr, INFO);
 } 
 /* do last works */
 unlink(ux_name);
 unlink(back_ux_name);
 exit(5);
}

static void exit_handler(int Code, void *arg)
{
 sprintf(mstr,"Exited With Code %d",Code);
 mess(mstr, INFO);
}

/* This functions removes a link, pointed to by accessor, from the designated 
   linked list, but does not free its memory.
   'role' can be one of WANT_ACCESS, PENDING and ACCESSOR.
   'type' can be one of SHM_READ and SHM_WRITE */
static void rem_accessor(int role, int type, struct accessor *accessor)
{
 (accessor->prev)->next = accessor->next;
 if(accessor->next)
  (accessor->next)->prev = accessor->prev;
 if(accessors_last[role][type] == accessor)
  accessors_last[role][type] = accessors_last[role][type]->prev;
}


/* add the link to the specified queue */
static void add_accessor(int role, int type, struct accessor *accessor)
{
 accessor->next = NULL;
 accessor->prev = accessors_last[role][type];
 accessors_last[role][type]->next = accessor;
 accessors_last[role][type] = accessor;
}


/* this function removes a link, poited to by accessor, from the designated
   linked list and then frees its memory. */
static void del_accessor(int role, int type, struct accessor *accessor)
{
 rem_accessor(role, type, accessor);
 if(role == WANT_ACCESS && type == SHM_READ && accessor->next &&
    SAME_ADDRESS(&(accessor->address), MY_ADDRESS))
 {
  /* if the above conditions hold, then we tried to make the owner machine 
     a reader to satisfy another machine's request. If we are deleting this 
     request, then better delete the original one as well */
  rem_accessor(role, type, accessor->next);
  free(accessor->next);
 }
 free(accessor);
}


/* check to see if there are any machines that have not sent aknowledgement
   for shm_ak_timeout seconds */
static void check_timeout(void)
{
 int count;
 struct accessor *pending, *pending2;
 static time_t last_check = 0;
 time_t curr_time = time(NULL);
 
 /* this routine may be called at intervals less than shm_ak_timeout */
 if(shm_ak_timeout <= 0 || global_timeout == 0) return;
 
 if(curr_time - last_check < shm_ak_timeout) return;
 for(count = 0; count < 2; count++)
 {
  pending = accessors_first[PENDING][count].next;
  while(pending)
  {
   if(curr_time - pending->time > shm_ak_timeout)
   {
    sprintf(mstr,"ERROR: Address: %s failed to Ak %s in time",
           inet_ntoa(pending->address.sin_addr), 
           (count == 0) ? "Reading" : "Writing");
    mess(mstr, INFO);
    pending2 = pending->prev;
    /* this is how we deal with processes who 
       don't keep their promises :-) */
    del_accessor(PENDING, count, pending);
    pending = pending2;
   } 
   pending = pending->next;
  }
 }
 last_check = curr_time;
}


/* uses an address to find an accessor.
   shm_man deals with machines in the cluster, not individual processes */ 
static struct accessor *find_accessor(int role, int type, 
                                                  struct sockaddr_in address)
{
 struct accessor *accessor;
 
 accessor = accessors_first[role][type].next;
 while(accessor)
 {
  if(SAME_MACHINE(&(accessor->address), address))
   break;
  accessor = accessor->next;
 }
 return accessor;
} 


/* clear it all. Used for example when a new writer forces all previous 
   readers to resign. */
static void del_all_accessors(int role, int type)
{
 struct accessor *accessor, *accessor2;
 
 accessor = accessors_first[role][type].next;
 while(accessor)
 {
  accessor2 = accessor->next;
  free(accessor);
  accessor = accessor2;
 }
 accessors_first[role][type].next = NULL;
 accessors_last[role][type] = &accessors_first[role][type];
}  


/* takes a request and changes it to a pending or accessor */
static void change_role(int new_role, int type, struct accessor *accessor)
{
 /* remove from the queue of the old role */
 if(new_role > WANT_ACCESS)
  rem_accessor(new_role - 1, type, accessor);

 /* add it to new role */
 if(SAME_ADDRESS(&(accessor->address), MY_ADDRESS))
 {
  accessor->next = accessors_first[new_role][type].next;;
  accessor->prev = &accessors_first[new_role][type];
  if(accessors_first[new_role][type].next)
   (accessors_first[new_role][type].next)->prev = accessor;
  accessors_first[new_role][type].next = accessor;
  if(accessors_last[new_role][type] == &accessors_first[new_role][type])
   accessors_last[new_role][type] = accessor;
 }
 else
 { 
  accessor->next = NULL;
  accessor->prev = accessors_last[new_role][type];
  accessors_last[new_role][type]->next = accessor;
  accessors_last[new_role][type] = accessor; 
 }
 accessor->time = time(NULL);
}


/* A brand new request for read or wirte has arrived. Put it in the 
   corresponding WANT_ACCESS queue. Won't add duplicate request from the same 
   computer. */
static BOOLEAN make_want_access(struct sockaddr_in address, int Pid, int type)
{
 struct accessor *want_access;

 if(verbose)
 {
  sprintf(mstr,"New access request from %s to %s",
          inet_ntoa(address.sin_addr), (type == 0) ? "Read" : "Write"); 
  mess(mstr, INFO); 
 }
 want_access = accessors_first[WANT_ACCESS][type].next;
 while(want_access)
 {
  if(SAME_ADDRESS(&(want_access->address), address))
   break;
  want_access = want_access->next;
 }
 if(want_access)
 {
  sprintf(mstr,"ERROR: Duplicate access request");
  mess(mstr, INFO);
  return FAIL;
 }
 want_access = (struct accessor *) malloc(sizeof(struct accessor));
 if(!want_access)
 {
  sprintf(mstr,"ERROR: No Mem");
  mess(mstr, INFO);
  return FAIL;
 }
 want_access->address = address;
 want_access->pid = Pid;
 /* add it */
 change_role(WANT_ACCESS, type, want_access); 
 return SUCCESS;
}


/* Allows a requesting machine to read or write. This is used after an 
   aknowledgement has arrived. */
static BOOLEAN allow(struct sockaddr_in address, int Pid, int op)
{
 struct message message;
 
 message.address = MY_ADDRESS;
 message.request = RES_SHMMAN;
 message.pid = Pid;
 message.info.key = my_shm.key;
 message.info.arg1 = my_shm.key;
 message.info.func = op;
 message.info.type = SHM;
 message.info.dsize = 0;
 
 if(send_in_message(&message, address) == FAIL)
 {
  sprintf(mstr,"ERROR: Can't Allow address %s to %s",
          inet_ntoa(address.sin_addr), (op == 0) ? "Read" : "Write"); 
  mess(mstr, INFO); 
  return FAIL;
 }
 if(verbose)
 {
  sprintf(mstr,"Allow address %s, pid %d to %s",
          inet_ntoa(address.sin_addr), Pid, (op == 0) ? "Read" : "Write"); 
  mess(mstr, INFO);
 }
 last_change = times(NULL);
 return SUCCESS;
}


/* when an aknowledge from a machine has arrived, it shows that, that 
   machine it ready to restart the stopped processes. allow it */ 
static BOOLEAN acknowledge(struct message *message)
{
 int res;
 struct accessor *pending; 
 
 pending = accessors_first[PENDING][SHM_WRITE].next;
 if(pending)
 {
  if(!SAME_ADDRESS(&(pending->address), message->address))
  {
   sprintf(mstr,"ERROR: Ak expected from %s, but came from %s",
     inet_ntoa(pending->address.sin_addr),inet_ntoa(message->address.sin_addr) );
   mess(mstr, INFO);
   return FAIL;
  }
  if(verbose)
  {
   sprintf(mstr,"Ak from a new writer");
   mess(mstr, INFO);
  }
  res = allow(message->address, pending->pid, WANTWR);
  if(res == SUCCESS)
  {
   del_all_accessors(ACCESSOR, SHM_READ);
   del_all_accessors(ACCESSOR, SHM_WRITE);
   change_role(ACCESSOR, SHM_WRITE, pending); 
  } 
  return res; 
 } 
 else 
 {
  /* the ak may be for a reader */
  pending = accessors_first[PENDING][SHM_READ].next; 
  while(pending)
  {
   if(verbose)
   {
    sprintf(mstr,"Ak from a new reader");
    mess(mstr, INFO);
   } 
   if(SAME_ADDRESS(&(message->address),pending->address))
    break; 
   pending = pending->next;
  }
  if(pending == NULL)
  {
   sprintf(mstr,"ERROR: Request For Unknown Reader");
   mess(mstr, INFO);
   return FAIL;
  } 
  res = allow(message->address, pending->pid, WANTRD);
  if(res == SUCCESS)
  {
   del_all_accessors(ACCESSOR, SHM_WRITE);
   change_role(ACCESSOR, SHM_READ, pending);
  } 
  return res; 
 } /* Ak for a reader */
 return SUCCESS;
}



/* tries to make a new machine the writer of the shared memory.
   It can't do this if:
   * There are no processes wanting to write.
   * There are some pending requests for read or write */
static BOOLEAN make_writer(void)
{
 BOOLEAN res;
 struct message message;
 struct accessor *reader, *reader2, *prev_reader, *current_writer;
 struct accessor *want_writer;
 
 want_writer = accessors_first[WANT_ACCESS][SHM_WRITE].next;
 if(!want_writer) return FAIL;
 
 /* no schedule if there are pending writers or readers */
 if(accessors_first[PENDING][SHM_WRITE].next != NULL) return FAIL;
 if(accessors_first[PENDING][SHM_READ].next != NULL) return FAIL;
  
 message.address = MY_ADDRESS;
 message.address2 = want_writer->address;
 message.info.arg1 = my_shm.size;
 message.info.type = SHM;
 message.info.key = my_shm.key;
 message.request = REQ_DOWORK;
 
 current_writer = accessors_first[ACCESSOR][SHM_WRITE].next;
 if(current_writer != NULL)
 {
  if(verbose)
  {
   sprintf(mstr,"one writer, and another wants to write");
   mess(mstr, INFO);
  }  
  if(find_accessor(ACCESSOR, SHM_WRITE, want_writer->address))
  {
   sprintf(mstr,"WARNING: Writer %s Wants to Write Key %d Again", 
           inet_ntoa(current_writer->address.sin_addr), my_shm.key);
   mess(mstr, INFO);        
   res = allow(want_writer->address, want_writer->pid, WANTWR);
   del_accessor(WANT_ACCESS, SHM_WRITE, want_writer); /* discard it */
   return res; 
  }
  else
  {
   message.info.func = RDPROT_SND;
   if(send_in_message(&message, current_writer->address) == FAIL)
   {
    sprintf(mstr,"ERROR: Can't Provide Contents");
    mess(mstr, INFO);
    del_accessor(WANT_ACCESS, SHM_WRITE, want_writer);
    return FAIL;
   }
   change_role(PENDING, SHM_WRITE, want_writer);
  }
 } 
 else
 {  /* there is no writer */ 
  /* there are no writers, handle all possible readers */
  if(verbose)
  {
   sprintf(mstr,"no writer, and one wants to write");
   mess(mstr, INFO);
  }
  prev_reader = find_accessor(ACCESSOR, SHM_READ, want_writer->address);
  if(prev_reader && verbose)
  {
   sprintf(mstr,"new writer at %s is a previous reader",
            inet_ntoa(prev_reader->address.sin_addr));
   mess(mstr, INFO);         
  }         
  reader = accessors_first[ACCESSOR][SHM_READ].next; 
  while(reader)
  {
   if(!prev_reader && SAME_ADDRESS(&(reader->address), MY_ADDRESS))
    message.info.func = RDPROT_SND;  
   else 
    message.info.func = RDPROT;
   if(prev_reader != reader) /* don't bother the future writer! */
   {
    if(verbose)
    {
     sprintf(mstr,"removing reader at %s",
             inet_ntoa(reader->address.sin_addr));
     mess(mstr, INFO);
    }        
    if(send_in_message(&message, reader->address) == FAIL)
    {
     sprintf(mstr,"ERROR: Can't read-protect");
     mess(mstr, INFO);
     reader2 = reader->prev;
     del_accessor(ACCESSOR, SHM_READ, reader);
     reader = reader2; 
     /* return FAIL; continue... */
    }         
   } 
   reader = reader->next; 
  }
  change_role(PENDING, SHM_WRITE, want_writer);
  if(prev_reader)
  {
   /* this machine has the contents, just allow it to continue */
   if(verbose)
   {
    sprintf(mstr,"A Previous Reader, Allowing It To Write");
    mess(mstr, INFO);
   }                  
   res = allow(want_writer->address, want_writer->pid, WANTWR);
   if(res == SUCCESS)
   {
    del_all_accessors(ACCESSOR, SHM_READ);
    del_all_accessors(ACCESSOR, SHM_WRITE);
    change_role(ACCESSOR, SHM_WRITE, want_writer);
   }
   else
    del_accessor(PENDING, SHM_WRITE, want_writer);
   return res; 
  }
 }  
 return SUCCESS;
}


/* tries to make a new machine the reader of the shared memory.
   It can't do this if:
   * There are no machines wanting to read.
   * There are other machines who want to write.
   * There are some pending requests for write */
static BOOLEAN make_reader(void)
{
 BOOLEAN res;
 struct accessor *want_reader;
 struct message message;
 struct accessor *owner_reader;
 struct accessor *current_writer;
 
 want_reader = accessors_first[WANT_ACCESS][SHM_READ].next;
 if(!want_reader) return FAIL;
 
 if(accessors_first[WANT_ACCESS][SHM_WRITE].next) return FAIL;
 if(accessors_first[PENDING][SHM_WRITE].next != NULL) return FAIL;
 
 if( !SAME_MACHINE(&want_reader->address, MY_ADDRESS) &&
     !find_accessor(ACCESSOR, SHM_WRITE, MY_ADDRESS) &&
     !find_accessor(ACCESSOR, SHM_READ, MY_ADDRESS))
 {    
  /* someone wants to read, but I don't have the contents, so make 
     me a reader too! I'll come back to the original request the next time 
     this function is called. But what happens if by that time a new machine 
     wants to write this shared memory? */
  if(make_want_access(MY_ADDRESS, 0, SHM_READ) == FAIL)
  {
   sprintf(mstr,"ERROR: Can not make the owner a candidate to read");
   mess(mstr, INFO);
   return FAIL;
  }
  want_reader = accessors_first[WANT_ACCESS][SHM_READ].next;      
 } 
 message.request = REQ_DOWORK;
 message.address = MY_ADDRESS;
 message.address2 = want_reader->address;
 message.info.key = my_shm.key;
 message.info.arg1 = my_shm.size;
 message.info.type = SHM;
 
 current_writer = accessors_first[ACCESSOR][SHM_WRITE].next;
 if(current_writer != NULL)
 {
  if(verbose)
  {
   sprintf(mstr,"one writer, and another wants to read");
   mess(mstr, INFO);
  } 
  if(SAME_MACHINE(&(want_reader->address), current_writer->address))
  {
   sprintf(mstr,"WARNING: Writer %s Wants To Read Key %d",
           inet_ntoa(current_writer->address.sin_addr), my_shm.key);
   mess(mstr, INFO);
   res = allow(want_reader->address, want_reader->pid, WANTRD);
   del_accessor(WANT_ACCESS, SHM_READ, want_reader);
   return res;
  }
  else
  {
   message.info.func = WRPROT_SND; /* writer will send the contents */
   if(send_in_message(&message, current_writer->address) == FAIL)
   {
    sprintf(mstr,"ERROR: Can't Force To Write Protect");
    mess(mstr, INFO);
    del_accessor(WANT_ACCESS, SHM_READ, want_reader);
    return FAIL;
   }
   /* the previous writer is a reader now */
   rem_accessor(ACCESSOR, SHM_WRITE, current_writer);
   add_accessor(ACCESSOR, SHM_READ, current_writer);
   if(verbose)
   {
    sprintf(mstr,"write protect request sent to writer");
    mess(mstr, INFO);
   } 
   change_role(PENDING, SHM_READ, want_reader); 
  }
 } /* if there is a writer */ 
 else
 { /* no writer, just add it to readers and send it the contents */
  if(verbose)
  {
   sprintf(mstr,"no writer, and one wants to read");
   mess(mstr, INFO);
  }
  if(find_accessor(ACCESSOR, SHM_READ, want_reader->address))
  {
   sprintf(mstr,"WARNING: Reader %s Wants To Read Key %d again",
           inet_ntoa(want_reader->address.sin_addr), my_shm.key);
   mess(mstr, INFO);
   res = allow(want_reader->address, want_reader->pid, WANTRD);
   del_accessor(WANT_ACCESS, SHM_READ, want_reader);
   return res;
  }
  else
  {
   message.info.func = NOP_SND;
   owner_reader = find_accessor(ACCESSOR, SHM_READ, MY_ADDRESS);
   if(owner_reader == NULL)
   {
    sprintf(mstr,"ERROR: owner Machine Is Not A Reader");
    mess(mstr, INFO);
    del_accessor(WANT_ACCESS, SHM_READ, want_reader);
    return FAIL;
   }
   if(send_in_message(&message, owner_reader->address) == FAIL)
   {
    sprintf(mstr,"ERROR: Can't Provide Contents");
    mess(mstr, INFO);
    del_accessor(WANT_ACCESS, SHM_READ, want_reader);
    return FAIL;
   } 
   change_role(PENDING, SHM_READ, want_reader);
  }
 }
 return SUCCESS;
}


/* see what the incomming message wants to say */
static void work_on(int ux_sock, struct message *message)
{
 if(message->info.key == my_shm.key)
 {
  switch(message->info.func)
  {
   case WANTWR:
    if(verbose)
    { 
     sprintf(mstr,"WANTWR from %s", inet_ntoa(message->address.sin_addr));
     mess(mstr, INFO);
    }
    make_want_access(message->address, message->pid, SHM_WRITE);
    break;
  
   case WANTRD: 
    if(verbose)
    {
     sprintf(mstr,"WANTRD from %s", inet_ntoa(message->address.sin_addr));
     mess(mstr, INFO);
    }
    make_want_access(message->address, message->pid, SHM_READ);
    break;
  
   case AK:
    if(verbose)
    {
     sprintf(mstr,"AK from %s", inet_ntoa(message->address.sin_addr));
     mess(mstr, INFO);
    }
    acknowledge(message);
    break;
  
   case DETACH:
    if(verbose)
    {
     sprintf(mstr,"DETACH from %s", inet_ntoa(message->address.sin_addr));
     mess(mstr, INFO);
    }
    if(shmdt(my_shm.attach) < 0)
    {
     sprintf(mstr,"ERROR: shmdt() failed  BECAUSE %s",strerror(errno));
     mess(mstr, INFO);
    }
    unlink(ux_name);
    unlink(back_ux_name);
    exit(0);
    break;  
   default:
    sprintf(mstr,"ERROR: Unknown Request from %s",
            inet_ntoa(message->address.sin_addr));
    mess(mstr, INFO);
    break;
  }
 }
 else
 {
  sprintf(mstr,"ERROR: From %s: Key %d Is Not Mine (%d)",
       inet_ntoa(message->address.sin_addr), message->info.key, my_shm.key);
  mess(mstr, INFO);
 }
}


static void return_result(struct dipc_info *di, char *data)
{
 void *pt[2];
 
 pt[0] = di;
 pt[1] = data;
 if(dipc(di->pid, DIPC_RETWORK, (int *)pt) < DIPC_SUCCESS)
 {
  sprintf(mstr,"ERROR: DIPC_RETWORK failed");
  mess(mstr, INFO);
 }
}

/* the back_door control center. It receives request from the UNIX domain 
   socket and executes them. This is for 'out of band' control of the 
   shm_man,  without the need to redesign it. The work here should be done 
   quickly, because while the shm_man is handling back_door requests, it 
   won't be able to handle normal, more important requests. */
static void bd_work_on(int ux_sockfd, struct back_door_message *bd_message)
{
 struct accessor *ac;
 struct back_door_message bd_response;
 int type1, type2, want_trans, act_trans;
 
 switch(bd_message->cmd)
 {
  case BACK_DOOR_INFO:
   if(verbose)
   { 
    sprintf(mstr,"back door: request for information\n");
    mess(mstr, INFO);
   } 
   /* display the information for shared memories, message queues, and
      semaphore sets */
   for(type1 = 0; type1 < 3; type1++)
   {
    for(type2 = 0; type2 < 2; type2++)
    {
     ac = accessors_first[type1][type2].next;
     while(ac)
     {
      bd_response.arg1 = type1; /* WANT_ACCESS, PENDING, ACCESOR */
      bd_response.arg2 = type2; /* SHM_READ, SHM_WRITE */
      bd_response.arg3 = 0;
      bd_response.arg4 = 0;
      bd_response.arg5 = ac->address.sin_addr.s_addr;
      want_trans = sizeof(bd_response);
      act_trans = writen(ux_sockfd, (char *)&bd_response, want_trans);
      if(act_trans < want_trans)
      {
       sprintf(mstr,"ERROR: sent %d From %d Bytes  BECAUSE %s",
            act_trans, want_trans,strerror(errno));
       mess(mstr, INFO);
       break;        
      }
      ac = ac->next;
     }
     if(ac != NULL) break;
    }
    if(ac != NULL) break;
   }
   if(type1 == 3)
   {
    /* this signals the end of information */ 
    bd_response.arg1 = -1;
    want_trans = sizeof(bd_response);
    act_trans = writen(ux_sockfd, (char *)&bd_response, want_trans);
    if(act_trans < want_trans)
    {
     sprintf(mstr,"ERROR: sent %d From %d Bytes  BECAUSE %s",
           act_trans, want_trans,strerror(errno));
     mess(mstr, INFO);        
    }
   }
   break;
      
  default:
   sprintf(mstr,"ERROR: Unknown back door request %d\n",bd_message->cmd);
   mess(mstr, INFO);
 }
}


void shm_man(struct dipc_info *di, char *data)
{
 int timeout;
 int count, count2;
 struct timeval TimeVal;
 fd_set rSelFD;
 int ux_newsockfd;
 int want_trans, act_trans;
 int ux_client_len, ux_len;
 struct sockaddr_un ux_client_addr, ux_addr;
 struct message message;
 struct sigaction sact;
 struct sockaddr_un back_ux_addr, back_ux_client_addr;
 int back_ux_len, back_ux_client_len;
 int back_ux_sockfd, back_ux_newsockfd;
 struct back_door_message bd_message;

 int key = di->key;
 struct accessor *first_writer;
 
 if(verbose)
 {
  sprintf(mstr,"shm_man Started: key %d",key);
  mess(mstr, INFO);
 } 
 sact.sa_handler = sig_handler;
 sigemptyset(&sact.sa_mask);
 sact.sa_flags = 0;
 
 if(sigaction(SIGINT, &sact, NULL) < 0)
 {
  sprintf(mstr,"ERROR: Can't Catch Signal  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
 } 
 if(sigaction(SIGTERM, &sact, NULL) < 0)
 {
  sprintf(mstr,"ERROR: Can't Catch Signal  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
 }
 if(sigaction(SIGSEGV, &sact, NULL) < 0)
 {
  sprintf(mstr,"ERROR: Can't Catch Signal BECAUSE %s",strerror(errno));
  mess(mstr,INFO);
 } 
 if(on_exit(exit_handler, NULL) < 0)
 {
  sprintf(mstr,"ERROR: Can't Catch Exit  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
 } 

 /* prepare the "back door" socket */
 sprintf(back_ux_name,"%s%s.%d",UNIX_PATH,SHM_BACKDOOR, key);
 bzero((char *)&back_ux_addr,sizeof(back_ux_addr));
 back_ux_addr.sun_family = AF_UNIX;
 strcpy(back_ux_addr.sun_path, back_ux_name);
 back_ux_len = strlen(back_ux_addr.sun_path) + sizeof(back_ux_addr.sun_family); 

 unlink(back_ux_name);
 back_ux_sockfd = plug(AF_UNIX, (struct sockaddr *)&back_ux_addr, back_ux_len);
 if(back_ux_sockfd < 0)
 { 
  sprintf(mstr,"ERROR: Can't Create Socket %s",back_ux_name);
  mess(mstr, INFO);
  unlink(back_ux_name);
  di->arg1 = -errno;
  return_result(di, data);
  exit(20);
 }
 
 sprintf(ux_name,"%sDIPC_SHM.%d",UNIX_PATH,di->key);  
 bzero((char *)&ux_addr,sizeof(ux_addr));
 ux_addr.sun_family = AF_UNIX;
 strcpy(ux_addr.sun_path,ux_name);
 ux_len = strlen(ux_addr.sun_path) + sizeof(ux_addr.sun_family);

 unlink(ux_name);
 ux_sockfd = plug(AF_UNIX, (struct sockaddr *) &ux_addr, ux_len);
 if(ux_sockfd < 0)
 {
  sprintf(mstr,"ERROR: Can't Create Socket %s",ux_name);
  mess(mstr, INFO);
  unlink(back_ux_name);
  di->arg1 = -errno;
  return_result(di, data);
  exit(20);
 }
 
 ticks_sec = (double) sysconf(_SC_CLK_TCK);
 my_shm.key = key;
 my_shm.id = shmget(key, 0, IPC_DIPC);
 if(my_shm.id < 0)
 {
  sprintf(mstr,"ERROR: shmget() failed  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
  unlink(ux_name);
  unlink(back_ux_name);
  di->arg1 = -errno;
  return_result(di, data);
  exit(20);
 }
 my_shm.attach = shmat(my_shm.id, NULL, 0);
 if(my_shm.attach == (char *) -1)
 {
  sprintf(mstr,"ERROR: shmat() failed  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
  unlink(ux_name);
  unlink(back_ux_name);
  di->arg1 = -errno;
  return_result(di, data);
  exit(20);
 } 
 my_shm.size = di->arg2;
 
 for(count = 0; count < 3; count++) 
  for(count2 = 0; count2 < 2; count2++)
  {
   accessors_first[count][count2].next = NULL;
   accessors_first[count][count2].prev = NULL;
   accessors_last[count][count2] = &accessors_first[count][count2];
  }
 
 /* this machine is the first writer. we have gradually make it so */
 make_want_access(MY_ADDRESS, 0, SHM_WRITE); 
 first_writer = accessors_first[WANT_ACCESS][SHM_WRITE].next;
 change_role(PENDING, SHM_WRITE, first_writer);
 change_role(ACCESSOR, SHM_WRITE, first_writer);
 
 /* continue where the employer left.
    this should have been done by the employer that started the shm_man. 
    This is done here, so as to make sure the shm_man is all setup and 
    ready to function. In this way when the original process creating 
    the shared memory is restarted, there is a manager ready to handle 
    the requests */
 return_result(di, data);
 
 while(1)
 {
  timeout = 0;
  TimeVal.tv_sec = 0;
  TimeVal.tv_usec = shm_hold_timeout * 1000000; 
  FD_ZERO(&rSelFD);
  FD_SET(ux_sockfd,&rSelFD);
  FD_SET(back_ux_sockfd, &rSelFD);
  
  if(select(FD_SETSIZE, &rSelFD, NULL, NULL, &TimeVal) < 0)
  {
   sprintf(mstr,"ERROR: select() failed  BECAUSE %s",strerror(errno));
   mess(mstr, INFO);
   unlink(ux_name);
   unlink(back_ux_name);
   exit(20);
  }
  if(FD_ISSET(ux_sockfd,&rSelFD))
  {
   if(verbose)
   {
    sprintf(mstr,"Call from UNIX socket");
    mess(mstr, INFO);
   }  
   ux_client_len = sizeof(ux_client_addr);
   if((ux_newsockfd = accept(ux_sockfd, (struct sockaddr *)&ux_client_addr,
                                                       &ux_client_len)) < 0)
   {
    sprintf(mstr,"ERROR: accept() Failed  BECAUSE %s",strerror(errno));
    mess(mstr, INFO);
    close(ux_sockfd);
    unlink(ux_name);
    unlink(back_ux_name);
    exit(20);
   } 
   want_trans = sizeof(message);
   act_trans = readn(ux_newsockfd,(char *)&message, want_trans);
   if(act_trans < want_trans)
   {
    sprintf(mstr,"ERROR: Received %d From %d Bytes  BECAUSE %s",
            act_trans, want_trans,strerror(errno));
    mess(mstr, INFO);        
   }
   else
    work_on(ux_newsockfd, &message);
   close(ux_newsockfd);
  }
  else if(FD_ISSET(back_ux_sockfd,&rSelFD))
  {
   if(verbose)
   {
    sprintf(mstr,"back door call");
    mess(mstr, INFO);
   }  
   back_ux_client_len = sizeof(back_ux_client_addr);
   if((back_ux_newsockfd = accept(back_ux_sockfd, 
         (struct sockaddr *)&back_ux_client_addr, &back_ux_client_len)) < 0)
   {
    sprintf(mstr,"ERROR: accept() Failed  BECAUSE %s",strerror(errno));
    mess(mstr, INFO);
    close(back_ux_newsockfd);
   }
   else
   {
    want_trans = sizeof(bd_message);
    act_trans = readn(back_ux_newsockfd,(char *)&bd_message, want_trans);
    if(act_trans < want_trans)
    {
     sprintf(mstr,"ERROR: Received %d From %d Bytes  BECAUSE %s",
            act_trans, want_trans,strerror(errno));
     mess(mstr, INFO);        
    }
    else
     bd_work_on(back_ux_newsockfd, &bd_message);
    close(back_ux_newsockfd);
   } 
  } /* if back door */
  else
   timeout = 1;
  /* timeout variable is used to prevent the unlikely event of a
     wrap-around in times().
     Either a time out, or a new request arrived.
     see if can schedule a new reader or witer */
  if(timeout || ((times(NULL) - last_change) / ticks_sec) > shm_hold_timeout)
  {
   check_timeout();
   /* first the writers, then the readers */
   if(make_writer() == FAIL)
    make_reader();
  }
 } /* while */  
}
