/*
 * worker.c
 *
 * part of dipcd source code
 *
 * Copyright (C) Kamran Karimi
 */

#include "dipcd.h"
#include "funcs.h"

/* this is forked by the fron_end. it receives a file descriptor, which
   it uses to exchange information with the remote task that caused its
   creation. Worker executes the function it is supposed to do and sends 
   the replies back to the computer that requested it. The front_end on
   that remote computer will deliver the results to the waiting process */

extern struct sockaddr_in MY_ADDRESS, REFEREE_ADDRESS;
extern int verbose;

static char ipc_type[3][4] = 
{
 "SHM",
 "MSG",
 "SEM"
};


static void exit_handler(int Code, void *argv)
{
 sprintf(mstr,"Exited With Code %d",Code);
 mess(mstr,INFO);
}

static void sig_handler(int signo)
{
 if(verbose)
 {
  sprintf(mstr,"Quiting By Signal No. %d",signo);
  mess(mstr, INFO);
 } 
 /* do last works */
 exit(5);
}

/* handle a remove request from the referee. Do it as root. There is
   no acknowledgement here */
static void req_rmid(struct message *message)
{
 int id;

 if(message->info.type == MSG)
 {
  id = msgget(message->info.key, IPC_DIPC);
  if(id != -1)
  {
   if(msgctl(id, IPC_RMID, NULL) < 0)
   {
    sprintf(mstr,"ERROR: Can't remove message queue %d  BECAUSE %s",
            message->info.key,strerror(errno));
    mess(mstr,INFO);                 
   }  
   else if(verbose)
   {
    sprintf(mstr,"Removed message queue %d",message->info.key);
    mess(mstr,INFO);
   } 
  }  
  else
  {
   sprintf(mstr,"ERROR: Can't find message queue %d",message->info.key);  
   mess(mstr, INFO);
  } 
 }   
 else if(message->info.type == SHM)
 {
  id = shmget(message->info.key, 0, IPC_DIPC);
  if(id != -1)
  {
   if(shmctl(id, IPC_RMID, NULL) < 0)
   {
    sprintf(mstr,"ERROR: Can't remove shared memory segment %d  BECAUSE %s",
            message->info.key,strerror(errno));
    mess(mstr, INFO);                
   }   
   else if(verbose)
   {
    sprintf(mstr,"Removed shared memory segment %d",message->info.key);
    mess(mstr,INFO);
   }
  }  
  else
  {
   sprintf(mstr,"ERROR: Can't find shared memory segment %d",
                                                message->info.key);  
   mess(mstr,INFO);
  }
 }
 else if(message->info.type == SEM)
 {
  id = semget(message->info.key, 0, IPC_DIPC);
  if(id != -1)
  {
   if(semctl(id, 0, IPC_RMID, NULL) < 0)
   {
    sprintf(mstr,"ERROR: Can't remove semaphore set %d  BECAUSE %s",
            message->info.key,strerror(errno));        
    mess(mstr, INFO);
   } 
   else if(verbose)
   {
    sprintf(mstr,"Removed semaphore set %d",message->info.key); 
    mess(mstr, INFO);
   }
  }  
  else
  {
   sprintf(mstr,"ERROR: Can't find semaphore set %d",message->info.key);  
   mess(mstr, INFO);
  }
 }
}

/* execute an IPC remote procedure call */
static void do_RPC(int in_sock, struct message *message)
{
 char data[MAXDATA];
 int count, retval, id;
 int arg2, arg3;
 int rep_sock;
 struct usr_info uinfo;
 int want_trans, act_trans;
 union semun arg;
 int type = message->info.type;
 
 want_trans = sizeof(struct usr_info);
 act_trans = readn(in_sock, (char *)&uinfo, want_trans);
 if(act_trans < want_trans)
 {
  sprintf(mstr,"ERROR: Received %d From %d Bytes  BECAUSE %s",
          act_trans, want_trans,strerror(errno));
  mess(mstr, INFO);        
  exit(20);
 }
 if(verbose)
 {
  sprintf(mstr,"Remote User Is: %s",uinfo.name);
  mess(mstr, INFO);
 }
 if((message->info.func == SEMOP) ||
    ((message->info.func == SEMCTL) && (message->info.arg3 == IPC_SET)) ||
    ((message->info.func == SEMCTL) && (message->info.arg3 == SETALL)) ||
    ((message->info.func == MSGSND) ||
    ((message->info.func == MSGCTL) && (message->info.arg2 == IPC_SET)) ||
    ((message->info.func == SHMCTL) && (message->info.arg2 == IPC_SET))))
 {
  /* these functions have some data */
  want_trans = message->info.dsize;   
  act_trans = get_in_data(in_sock, data, message);
  if(act_trans < want_trans)
  {
   sprintf(mstr,"ERROR: Received %d From %d Bytes  BECAUSE %s",
           act_trans, want_trans,strerror(errno));
   mess(mstr, INFO);        
   exit(20);
  }
 } 
 if(verbose)
 {
  sprintf(mstr,"Required information for RPC received");
  mess(mstr, INFO);
 }
 /* no further need for in_sock */
 if(message->info.func == TSTGET)
 {
  arg2 = message->info.arg2;
  arg3 = message->info.arg3;
 }
 else
 {
  arg2 = 0;
  arg3 = IPC_DIPC;
 }
 /* continue as (become) the remote user */
 if(become(uinfo.name) == FAIL)
 {
  errno = EPERM;
  retval = -1; 
 }
 else
 {
  /* this is will be tested When we use xxxget() */  
  if(message->info.func == TSTGET)
  { 
   if(verbose)
   {
    sprintf(mstr,"TSTGET");
    mess(mstr, INFO);
   } 
  }
  if(type == SEM) 
   id = semget(message->info.arg1, arg2, arg3); 
  else if(type == MSG)  
   id = msgget(message->info.arg1, arg3);
  else 
   id = shmget(message->info.arg1, arg2, arg3);
  
  if(id < 0)
  {
   if(verbose)
   {
    sprintf(mstr,"Could not access IPC structure");
    mess(mstr, INFO);
   } 
   retval = id;
  }
  else switch(message->info.func)
  {
   case TSTGET:
    retval = id;
    break;
    
   case SEMCTL:
    if(verbose)
    {
     sprintf(mstr,"SEMCTL");
     mess(mstr, INFO);
    }
    switch(message->info.arg3)
    {
     case SETVAL:
      arg.val = message->info.arg4;
      break;
     
     case IPC_SET:
     case IPC_STAT:
      arg.buf = (struct semid_ds *) data;
      break;
      
     case SETALL:
     case GETALL:
      arg.array = (ushort *) data;
      break;
    }  
    retval = semctl(id, message->info.arg2, message->info.arg3, arg); 
    break;
   
   case SEMOP:
    if(verbose)
    {
     sprintf(mstr,"SEMOP");
     mess(mstr, INFO);
    }
    /* sorry, no SEM_UNDO flag! */
    for(count = 0; count < message->info.arg2; count++)
     (((struct sembuf *)data)[count]).sem_flg &= ~SEM_UNDO;
    retval = semop(id, (struct sembuf *)data, message->info.arg2);  
    break;
    
   case MSGCTL:
    if(verbose)
    {
     sprintf(mstr,"MSGCTL");
     mess(mstr, INFO);
    }
    retval = msgctl(id, message->info.arg2, (struct msqid_ds *) data); 
    break;
   
   case MSGSND:
    if(verbose)
    {
     sprintf(mstr,"MSGSND");
     mess(mstr, INFO);
    }
    retval = msgsnd(id, (struct msgbuf *) data, message->info.arg2, 
                                                  message->info.arg3);  
    break;
   
   case MSGRCV:
    if(verbose)
    {
     sprintf(mstr,"MSGRCV");
     mess(mstr, INFO);
    }
    retval = msgrcv(id, (struct msgbuf *)data, message->info.arg2, 
                             message->info.arg3, message->info.arg4);
    break; 
  
   case SHMCTL:
    if(verbose)
    {
     sprintf(mstr,"SHMCT");
     mess(mstr, INFO);
    } 
    retval = shmctl(id, message->info.arg2, (struct shmid_ds *) data); 
    break;
  
   default:
    sprintf(mstr,"ERROR: Unknown Function");
    mess(mstr, INFO);
    errno = EINVAL;
    retval = -1; 
    break;
  }
  if(become("root") == FAIL)
  {
   sprintf(mstr,"ERROR: Can't Become Root");
   mess(mstr, INFO);
  }
 }

 if(retval < 0) message->info.arg1 = -errno;
 else message->info.arg1 = retval;
 if(verbose)
 {
  sprintf(mstr,"Returning results. Return value is %d",message->info.arg1);
  if(retval < 0)
  {
   char temp[80];
   sprintf(temp, "  (%s)",strerror(errno));
   strcat(mstr, temp);
  }
  mess(mstr, INFO);
 }
 rep_sock = ring(AF_INET, (struct sockaddr *)&(message->address),
                                                sizeof(message->address));
 if(rep_sock < 0)
 {
  sprintf(mstr,"ERROR: Can't connect back  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
  exit(20);
 } 
 message->request = RES_DOWORK;
 message->address = MY_ADDRESS;
 message->arch = MY_DIPC_ARCH;
 want_trans = sizeof(*message);
 act_trans = writen(rep_sock, (char *)message, want_trans);
 if(act_trans < want_trans)
 {
  sprintf(mstr,"ERROR: Sent %d From %d Bytes  BECAUSE %s",
          act_trans, want_trans,strerror(errno));
  mess(mstr, INFO);        
  exit(20);
 }
 if(verbose)
 {
  sprintf(mstr,"Reply Message sent");
  mess(mstr, INFO);
 }

 if(((message->info.func == SEMCTL) && (message->info.arg2 == IPC_STAT)) ||
    ((message->info.func == SEMCTL) && (message->info.arg2 == GETALL))   ||
    ((message->info.func == MSGRCV) ||
    ((message->info.func == MSGCTL) && (message->info.arg2 == IPC_STAT)) ||
    ((message->info.func == SHMCTL) && (message->info.arg2 == IPC_STAT))))
 {
  /* these calls require some results to be sent back */
  want_trans = message->info.dsize;   
  act_trans = writen(rep_sock, data, want_trans);
  if(act_trans < want_trans)
  {
   sprintf(mstr,"ERROR: Sent %d From %d Bytes  BECAUSE %s",
           act_trans, want_trans,strerror(errno));
   mess(mstr, INFO);        
   exit(20);
  }
 } 
 close(rep_sock);
 if(verbose)
 {
  sprintf(mstr,"RPC done and exiting");
  mess(mstr, INFO);
 }
 exit(0);
}

/* transfer the shared memory contents to another computer */
static BOOLEAN shm_send_contents(struct message *message)
{
 char *attach;
 int in_sock;
 int id;
 int want_trans, act_trans;
 struct message response;

 id = shmget(message->info.key, 0, IPC_DIPC);
 if(id < 0)
 {
  sprintf(mstr,"ERROR: shmget() failed for key %d  BECAUSE %s",
          message->info.key,strerror(errno));
  mess(mstr, INFO);        
  return FAIL;
 }
 if(verbose)
 {
  sprintf(mstr,"shmget() was successful. Size = %d",message->info.arg1); 
  mess(mstr, INFO);
 }
 attach = (char *)shmat(id,0,0);
 if(attach == (char *) -1)
 { 
  sprintf(mstr,"ERROR: shmat() Failed  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
  return FAIL;
 }
 response = *message;
 response.arch = MY_DIPC_ARCH;
 response.address = MY_ADDRESS;
 response.address2 = message->address; /* owner */
 response.info.func = RCV;
 /* sender determines the size */
 if(verbose)
 {
  sprintf(mstr,"Sending contents to %s",inet_ntoa(message->address2.sin_addr));
  mess(mstr, INFO);
 }
 in_sock = ring(AF_INET, (struct sockaddr *)&(message->address2),
                                        sizeof(message->address2));
 if(in_sock < 0)
 {
  sprintf(mstr,"ERROR: connect() Failed  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
  shmdt(attach);
  return FAIL;
 }
 want_trans = sizeof(response);
 act_trans = writen(in_sock, (char *) &response, want_trans);
 if(act_trans < want_trans)
 {
  sprintf(mstr,"ERROR: Sent %d From %d Bytes  BECAUSE %s",
          act_trans, want_trans,strerror(errno));
  mess(mstr, INFO);
  shmdt(attach);         
  return FAIL;
 }
 want_trans = message->info.arg1; 
 act_trans = writen(in_sock, attach, want_trans); 
 if(act_trans < want_trans)
 {
  sprintf(mstr,"ERROR: Sent %d From %d Bytes  BECAUSE %s",
          act_trans, want_trans,strerror(errno));
  mess(mstr, INFO);
  shmdt(attach);        
  return FAIL;
 }
 shmdt(attach);
 return SUCCESS;
}


/* receive the shared memory contents. After the shared memory contents
   are successfully received, sends an acknowledge message to the shared
   memory manager */
static BOOLEAN shm_receive_contents(int in_sock, struct message *message)
{
 char *attach;
 int id;
 int want_trans, act_trans;
 struct message response;

 id = shmget(message->info.key, 0, IPC_DIPC);
 if(id < 0)
 {
  sprintf(mstr,"ERROR: shmget() Failed  BECASUE %s",strerror(errno));
  mess(mstr, INFO);
  return FAIL;
 }
 if(verbose)
 {
  sprintf(mstr,"shmget() was successful. Size = %d",message->info.arg1); 
  mess(mstr, INFO);
 }
 attach = (char *)shmat(id,0,0);
 if(attach == (char *) -1)
 { 
  sprintf(mstr,"ERROR: shmat() Failed  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
  return FAIL;
 }
 want_trans = message->info.arg1;
 act_trans = readn(in_sock, attach, want_trans);
 if(act_trans < want_trans)
 {
  sprintf(mstr,"ERROR: Received %d From %d Bytes  BECAUSE %s",
          act_trans, want_trans,strerror(errno));
  mess(mstr,INFO);
  shmdt(attach);
  return FAIL;
 }
 shmdt(attach);
 
 response = *message; 
 response.address = MY_ADDRESS;
 response.request = REQ_SHMMAN;
 response.info.func = AK;
 if(send_in_message(&response, message->address2) == FAIL)
 {
  sprintf(mstr,"ERROR: Can't send AK to %s",
                            inet_ntoa(message->address2.sin_addr));
  mess(mstr,INFO);
  return FAIL;
 }
 if(verbose)
 {
  sprintf(mstr,"sent AK to %s", inet_ntoa(message->address2.sin_addr));
  mess(mstr, INFO);
 } 
 return SUCCESS;                               
}

/* read protect a shared memory */
static void read_protect(key_t key)
{
 int id, arg;
 
 id = shmget(key, 0, IPC_DIPC);
 if(id < 0)
 {
  sprintf(mstr,"ERROR: shmget() Failed  BECASUE %s",strerror(errno));
  mess(mstr, INFO);
  exit(20);
 }
 shmctl(id, SHM_UNLOCK, NULL);
 arg = DISABLE_RD;
 if(dipc(key,DIPC_REMPROT,&arg) < DIPC_SUCCESS)
 {
  sprintf(mstr,"DIPC_REMPROT failed");
  mess(mstr, INFO);
 }
}


static void shm_dowork(int in_sock, struct message *message)
{
 int arg;
 
 if(verbose)
 {
  sprintf(mstr,"shm_dowork");
  mess(mstr, INFO);
 }
 switch(message->info.func)
 {
  /* first write protect, then send shm contents */  
  case WRPROT_SND:
   if(verbose)
   {
    sprintf(mstr,"WRPROT_SND");
    mess(mstr, INFO);
   }
   arg = DISABLE_WR;
   if(dipc(message->info.key,DIPC_REMPROT, &arg) < DIPC_SUCCESS)
   {
    sprintf(mstr,"ERROR: DIPC_REMPROT failed");
    mess(mstr, INFO);
   }
   shm_send_contents(message);  
   break;  
  
  /* just read-protect */
  case RDPROT:
   if(verbose)
   {
    sprintf(mstr,"RDPROT");
    mess(mstr, INFO);
   }
   read_protect(message->info.key); 
   break;
  
  /* first read protect, then send the contents */
  case RDPROT_SND:
   if(verbose)
   {
    sprintf(mstr,"RDPROT_SND");
    mess(mstr, INFO);
   }
   read_protect(message->info.key);
   shm_send_contents(message);  
   break; 
  
  case NOP_SND: /* just send it */
   if(verbose)
   {
    sprintf(mstr,"NOP_SND");
    mess(mstr, INFO);
   }
   shm_send_contents(message);  
   break;
   
  case RCV: /* receive it as a result of direct request */
   if(verbose)
   {
    sprintf(mstr,"RCV");
    mess(mstr, INFO);
   }
   shm_receive_contents(in_sock, message);  
   break; 
   
  default:
   sprintf(mstr,"ERROR: Unknown Function");
   mess(mstr, INFO);
   break; 
 }  
}

/* this routine handles requests comming from other machines */
void worker(int in_sock, struct message message)
{
 struct sigaction sact;

 if(verbose)
 {
  sprintf(mstr,"Started: Request %s, funcion: %s", 
     request_string(message.request), work_string(message.info.func));
  mess(mstr, INFO);
 }
 sact.sa_handler = sig_handler;
 sigemptyset(&sact.sa_mask);
 sact.sa_flags = 0;

 if(on_exit(exit_handler, NULL) < 0)
 {
  sprintf(mstr,"ERROR: Can't install exit handler BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
 } 

 if(sigaction(SIGINT, &sact, NULL) < 0)
 {
  sprintf(mstr,"ERROR: Can't Catch Signal  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
 } 
 if(sigaction(SIGTERM, &sact, NULL) < 0)
 {
  sprintf(mstr,"ERROR: Can't Catch Signal  BECAUSE %s",strerror(errno));
  mess(mstr, INFO);
 }
 if(sigaction(SIGSEGV, &sact, NULL) < 0)
 {
  sprintf(mstr,"ERROR: Can't Catch Signal BECAUSE %s",strerror(errno));
  mess(mstr,INFO);
 } 
 switch(message.info.func)
 {
  case RMID:
   if(verbose)
   {
    sprintf(mstr,"RMID  type %s",ipc_type[message.info.type]);
    mess(mstr,INFO);
   }
   req_rmid(&message);
   break;                 
  
  case SEMCTL:
  case SEMOP:
  case MSGCTL:
  case MSGSND:
  case MSGRCV:
  case SHMCTL:
  case TSTGET:
   do_RPC(in_sock, &message);
   break;
   
  default:
   if(verbose)
   {
    sprintf(mstr,"DOWORK  type %s", ipc_type[message.info.type]);
    mess(mstr, INFO);
   }
   if(message.info.type == SHM)
    shm_dowork(in_sock, &message); 
   else
   {
    sprintf(mstr,"ERROR: Unknown Type");
    mess(mstr, INFO);
    exit(20);
   }
   break;
 }
 close(in_sock);
 exit(0);
}
