/*
 * linux/ipc/dipc.c
 * 
 * Copyright (C) Kamran Karimi
 *
 */

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/ipc.h>
#include <linux/shm.h>
#include <linux/msg.h>
#include <linux/sem.h>
#include <linux/stat.h>
#include <linux/malloc.h>

#include <asm/segment.h>
#include <asm/pgtable.h>
#include <linux/dipc.h>

#define DIPC_MAX_WAIT 127
 
/* this structure holds the network address of the ipc owner */
static struct dipc_address
{
 int type; /* SHM, SEM or MSG */
 int key;
 long address; 
 struct dipc_address *next,*prev;
} dipc_address_first = { 0, 0, 0, NULL, NULL };
static struct dipc_address *dipc_address_last = &dipc_address_first;

static struct dipc_request dipc_request_first = { {0,0,0,0,0,0,0,0,0,0,0},0,
                                                 NULL, NULL, NULL, NULL };
static struct dipc_request *dipc_request_last = &dipc_request_first;

static struct wait_queue *dipc_man_lock = NULL;
static struct task_struct *dipc_task = NULL; /* dipc manager */
static int My_Address;

static struct dipc_wait_t 
{
 struct wait_queue *queue;
 int flag;
}dipc_wait[3][DIPC_MAX_WAIT];


void dipc_init (void)
{
 int count1, count2;

 for(count1 = 0; count1 < 3; count1++)
 {
  for (count2 = 0; count2 < DIPC_MAX_WAIT; count2++)
  {
   dipc_wait[count1][count2].queue = NULL;
   dipc_wait[count1][count2].flag = 0;
  }
 } 	
 printk("Distributed System V IPC.  Version %s\n", DIPC_VERSION);
}

int dipc_present(void)
{
 if(dipc_task != NULL)
  return 1;
 return 0;
}
 
/* if it is the current task, or is the child of the manager, return 1 */
int dipc_manager(struct task_struct *task)
{
 if(dipc_present())
 {
  if((task == dipc_task) || (task->p_opptr == dipc_task))
   return 1;
  /* see if this is a grand child */   
  if((task->p_opptr)->p_opptr &&
     ((task->p_opptr)->p_opptr == dipc_task)) 
   return 1;
 }      
 return 0;
}

void dipc_excl(key_t key, int type)
{
 key = key % DIPC_MAX_WAIT;
 /**printk("KERNEL EXCLUDE key %d, type %d, by task %d\n",key,type,current->pid);*/
 dipc_wait[type][key].flag = 1;
}

int dipc_check(key_t key, int type)
{
 key = key % DIPC_MAX_WAIT;
 
 if(dipc_wait[type][key].flag == 1)
 {
  /**printk("KERNEL STOPPED key %d, type %d, task %d\n",key,type,current->pid);*/
  interruptible_sleep_on(&dipc_wait[type][key].queue);
  return 1;
 }
 return 0; 
}

void dipc_wake(key_t key, int type)
{
 key = key % DIPC_MAX_WAIT;
 
 if(dipc_wait[type][key].flag == 1)
 {
  /**printk("KERNEL RESTART key %d, type %d, task %d\n",key,type,current->pid);*/
  dipc_wait[type][key].flag = 0;
  wake_up(&dipc_wait[type][key].queue);
 }
}


static int add_owner_address(key_t key, int type, long net_address)
{
 struct dipc_address *da;
 /**printk("Kernel: add_owner_address(%d, %d, %ld)\n", key, type, net_address);*/
 da = dipc_address_first.next;
 while(da)
 {
  if((da->key == key) && (da->type == type)) break;
  da = da->next;
 }
 if(da != NULL)
  printk("add_owner_address(%d, %d): adding duplicate owner\n", key, type); 
 da = kmalloc(sizeof(struct dipc_address), GFP_KERNEL);
 if(da == NULL) return -1;
 da->type = type;
 da->key = key;
 da->address = net_address;
 da->next = NULL;
 da->prev = dipc_address_last;
 dipc_address_last->next = da;
 dipc_address_last = da;
 return 0;
}


/* for use in dipcstat. This is called from shm.c, msg.c and sem.c */
long silent_find_owner_address(key_t key, int type)
{
 struct dipc_address *da;
 
 /**printk("Kernel: find_owner_address(%d, %d)\n",key, type);*/
 da = dipc_address_first.next;
 while(da)
 {
  if((da->key == key) && (da->type == type)) break;
  da = da->next;
 }
 if(da == NULL)
  return -1;
 else return da->address;
}


static long find_owner_address(key_t key, int type)
{
 struct dipc_address *da;
 
 /**printk("Kernel: find_owner_address(%d, %d)\n",key, type);*/
 da = dipc_address_first.next;
 while(da)
 {
  if((da->key == key) && (da->type == type)) break;
  da = da->next;
 }
 if(da == NULL)
 {
  printk("Kernel: where is the owner address to find?\n");
  return -1;
 }
 else return da->address;
}
 

static int del_owner_address(key_t key, int type)
{
 struct dipc_address *da;
 
 /**printk("Kernel: del_owner_address(%d, %d)\n",key, type);*/
 da = dipc_address_first.next;
 while(da)
 {
  if((da->key == key) && (da->type == type)) break;
  da = da->next;
 }
 if(da == NULL)
 {
  printk("Kernel: where is the address to delete?\n");
  return -1;
 } 
 else
 {
  if(dipc_address_last == da) 
   dipc_address_last = da->prev;
  (da->prev)->next = da->next;
  if(da->next)
   (da->next)->prev = da->prev;
  kfree(da);
  return 0;
 }
}



/* Tasks use this routine to announce their need to search
   for a new key of type "type". A manager should later handle it. */
int dipc_search_key(key_t key, int type, int distrib, int *size)
{
 struct dipc_request *dr;
 int result;
 
 /**printk("Kernel: dipc_search_for_key(%d, %d)\n",key, type);*/
 dr = kmalloc(sizeof(struct dipc_request), GFP_KERNEL);
 if(dr == NULL) return -ENOMEM; /* set errno */
 dr->info.key = key;
 dr->info.type = type;
 dr->info.pid = current->pid;
 dr->info.uid = current->uid;
 dr->info.func = SEARCH;
 dr->info.arg1 = key;
 dr->info.arg2 = distrib;
 dr->info.arg3 = 0;
 dr->info.arg4 = 0;
 dr->info.owner = -1;
 dr->working = 0;

 dr->next = dr->prev = NULL;
 dr->lock = NULL;
 dr->info.dsize = 0;
 dr->data = NULL;
 
 dipc_add_request(dr);

 if(dr->info.arg1 >= 0)
 {
  result = dr->info.arg2;
  *size = dr->info.arg3;
  add_owner_address(key, type, dr->info.arg4);
 }
 else
 {
  result = NOT_FOUND;
  *size = 0;
  add_owner_address(key, type, My_Address);
 }
 kfree(dr); 
 return result;
}


/* this routine tells if an attempt to get a new key succeded or not.
   if success is one, the result has some information about the key */ 
int dipc_commit_key(key_t key, int type, int size, int mode)
{
 struct dipc_request *dr;
 
 /**printk("Kernel: dipc_key_result(%d, %d)\n",key, type);*/
 dr = kmalloc(sizeof(struct dipc_request), GFP_KERNEL);
 if(dr == NULL) return -ENOMEM; /* set errno */
 dr->info.key = key;
 dr->info.type = type;
 dr->info.pid = current->pid;
 dr->info.uid = current->uid;
 dr->info.func = COMMIT;
 dr->info.arg1 = key;
 dr->info.arg2 = size;
 dr->info.arg3 = mode;
 dr->info.arg4 = 0;
 dr->info.owner = -1; 
 dr->working = 0;
 dr->next = dr->prev = NULL;
 dr->lock = NULL;
 dr->info.dsize = 0;
 dr->data = NULL;
 
 dipc_add_request(dr); /* ignores errors */

 if(size <= 0) /* failed */
  del_owner_address(key, type);
 kfree(dr); 
 return 0;
}

/* this routine tests to see if the key can be "gotten" in the 
   owner machine. This is for access checking. result is the id of the
   remote key, or error code */ 
int dipc_test_key(key_t key, int type, int size, int flag)
{
 struct dipc_request *dr;
 int result;
 
 /**printk("Kernel: dipc_test_key(%d, %d)\n",key, type);*/

 dr = kmalloc(sizeof(struct dipc_request), GFP_KERNEL);
 if(dr == NULL) return -ENOMEM; /* set errno */
 dr->info.key = key;
 dr->info.type = type;
 dr->info.pid = current->pid;
 dr->info.uid = current->uid;
 dr->info.func = TSTGET;
 dr->info.arg1 = key;
 dr->info.arg2 = size;
 dr->info.arg3 = flag;
 dr->info.arg4 = 0;
 dr->info.owner = -1; 
 dr->working = 0;
 dr->next = dr->prev = NULL;
 dr->lock = NULL;
 dr->info.dsize = 0;
 dr->data = NULL;
 
 dipc_add_request(dr);
 
 result = dr->info.arg1; /* any kind of error (network, test error...) */
 kfree(dr); 
 return result;
}



int dipc_rmid_key(key_t key, int type)
{
 struct dipc_request *dr;
 
 /**printk("Kernel: dipc_rmid_key(%d, %d)\n", key, type);*/
 
 dr = kmalloc(sizeof(struct dipc_request), GFP_KERNEL);
 if(dr == NULL) return -ENOMEM; /* set errno */
 dr->info.key = key;
 dr->info.type = type;
 dr->info.pid = current->pid;
 dr->info.uid = current->uid;
 dr->info.func = RMID;
 dr->info.arg1 = key;
 dr->info.arg2 = 0;
 dr->info.arg3 = 0;
 dr->info.arg4 = 0;
 dr->info.owner = -1;
 dr->working = 0;
 dr->next = dr->prev = NULL;
 dr->lock = NULL;
 dr->info.dsize = 0;
 dr->data = NULL;
 
 dipc_add_request(dr); /* ignores errors */

 kfree(dr); 
 return 0;
}



/* Tasks use this routine to announce that they are removing 
   the specified key. This should be reflected in the referee tables */
int dipc_del_key(key_t key, int type)
{
 struct dipc_request *dr;
 
 /**printk("Kernel: dipc_del_key(k %d, t %d, p %d)\n", key, type, current->pid);*/
 
 dr = kmalloc(sizeof(struct dipc_request), GFP_KERNEL);
 if(dr == NULL) return -ENOMEM; /* set errno */
 dr->info.key = key;
 dr->info.type = type;
 dr->info.pid = current->pid;
 dr->info.uid = current->uid;
 dr->info.func = DELKEY;
 dr->info.arg1 = key;
 dr->info.arg2 = 0;
 dr->info.arg3 = 0;
 dr->info.arg4 = 0;
 dr->info.owner = -1; 
 dr->working = 0;
 dr->next = dr->prev = NULL;
 dr->lock = NULL;
 dr->info.dsize = 0;
 dr->data = NULL;
 
 dipc_add_request(dr); /* ignores error */
 del_owner_address(key, type);
 kfree(dr); 
 return 0;
}



/* tasks call this to put dipc work for the manager */
void dipc_add_request(struct dipc_request *dr)
{
 long owner;
 
 /**printk("Kernel: dipc_add() func: %d  pid: %d\n", dr->info.func, dr->info.pid);*/
 if(dipc_task == NULL) 
 {
  dr->info.arg1 = -EIO;
  return;
 }
 if(dr->info.func != RMID && dr->info.func != DELKEY &&
    dr->info.func != SEARCH && dr->info.func != COMMIT)
 {
  owner = find_owner_address(dr->info.arg1, dr->info.type);
  if(owner < 0)
  { 
   printk("Kernel dipc_add: can not find owner: key %d,  type %d\n",
                                      dr->info.arg1, dr->info.type);  
   dr->info.arg1 = -EIO;
   return;
  }
  dr->info.owner = owner;  
 } 
 dr->next = NULL;
 dr->prev = dipc_request_last;
 dipc_request_last->next = dr;
 dipc_request_last = dr;
 
 wake_up(&dipc_man_lock);
 interruptible_sleep_on(&(dr->lock));
  
 /* see if we were woken for an interrupt */
 if(current->signal & ~current->blocked)
  dr->info.arg1 = -EINTR; 
  
 if(dipc_request_last == dr)
  dipc_request_last = dr->prev;
 (dr->prev)->next = dr->next;
 if(dr->next) 
  (dr->next)->prev = dr->prev; 
 return;
}


/* manager calls this function to collect work */
static int get_new_request(int type, void *ptr)
{
 struct dipc_request *dr;
 int err;
 void *pt[2];
 struct dipc_info *info;
 char *data;
 
/* printk("Kernel: get_dipc_request()\n"); */
 err = verify_area(VERIFY_READ, ptr, 2 * sizeof(void *));
 if(err) return err;
 memcpy_fromfs(pt, ptr, 2 * sizeof(void *));
 info = pt[0];
 data = pt[1];
 
 dr = dipc_request_first.next;
 while(dr)
 {
  if(dr->working == 0)
  {
   /* if type == 0, return any type */
   if(type == 0) break;
   else if(type == dr->info.type) break;
  }
  dr = dr->next;
 }
 if(dr)
 {
  dr->working = 1;
  err = verify_area(VERIFY_WRITE, info, sizeof(struct dipc_info));
  if(err) return err;
  memcpy_tofs(info, &dr->info, sizeof(struct dipc_info));
  if(dr->info.dsize)
  {
   err = verify_area(VERIFY_WRITE, data, dr->info.dsize);
   if(err) return err;
   memcpy_tofs(data, dr->data, dr->info.dsize);
  }
  return DIPC_SUCCESS; 
 }
 /* nothing suitable found */
 return DIPC_SUCCESS + 1;
} 


/* this routine returns the result of a request, hopefully after
   successful execution at the remote owner */
static int put_result_of_request(int pid, void *ptr)
{
 struct dipc_request *dr;
 int err;
 void *pt[2];
 struct dipc_info *info;
 char *data;
 
 /**printk("Kernel: ret_dipc_request(): task %d\n", pid);*/
 err = verify_area(VERIFY_READ, ptr, 2 * sizeof(void *));
 if(err) return err;
 memcpy_fromfs(pt, ptr, 2 * sizeof(void *));
 info = pt[0];
 data = pt[1];
 dr = dipc_request_first.next;
 while(dr)
 {
  if(dr->info.pid == pid) break;
  dr = dr->next;
 }
 if(dr)
 {
  /* save the pointers before over write */
  err = verify_area(VERIFY_READ, info, sizeof(struct dipc_info));
  if(err) return err;
  memcpy_fromfs(&(dr->info), info, sizeof(struct dipc_info));
  if((dr->info.arg1 >= 0) && dr->info.dsize)
  {
   err = verify_area(VERIFY_READ, data, dr->info.dsize);
   if(err) return err;
   memcpy_fromfs(dr->data, data, dr->info.dsize);
  }
  /* Now wake it up */ 
  wake_up(&(dr->lock));
  return DIPC_SUCCESS;  
 } /* if dr */  
 printk("Kernel: where is the dipc request?\n");
 return DIPC_FAIL; 
}


static int wait_for_work(void)
{
 struct dipc_request *dr;
  
 /* get_dipc_request */
 dr = dipc_request_first.next;
 while(dr)
 {
  if(dr->working == 0) break;
  dr = dr->next;
 }
 if(dr != NULL)
  return DIPC_SUCCESS;
 
 interruptible_sleep_on(&dipc_man_lock);
 
 return DIPC_SUCCESS;
}

static int dipc_cleanup(void)
{
 struct dipc_request *dr;
 
 dipc_task = NULL;
 dr = dipc_request_first.next;
 while(dr)
 {
  dr->info.arg1 = -EIO;
  wake_up(&dr->lock);
  dr = dr->next;
 }
 dipc_request_first.next = dipc_request_first.prev = NULL;
 dipc_request_last = &dipc_request_first;
 
 return DIPC_SUCCESS;
}


extern int get_msg_status(int key, struct msqid_ds *ptr);
extern int get_shm_status(int key, struct shmid_ds *ptr);
extern int get_sem_status(int key, struct semid_ds *ptr); 

/* in shm.c */
extern int disable_prot(key_t key, int cmd, int *mask);  
extern int shm_reader(void);
extern int shm_writer(void);
 
 
int sys_dipc(int first, int cmd, void *ptr)
{
 switch(cmd)
 {
  case DIPC_GETWORK:
   return get_new_request(first, ptr); 
   break; 
  
  case DIPC_RETWORK:
   return put_result_of_request(first, ptr);
   break; 
 
  case DIPC_SLEEP:
   return wait_for_work(); 
   break;
   
  case DIPC_SETMNG:
   dipc_task = current;
   My_Address = first; /* ints are 32 bits */
   return DIPC_SUCCESS;
   break;
   
  case DIPC_REMMNG:
   return dipc_cleanup();
   break;

  case DIPC_REMPROT:
   return disable_prot(first, cmd, ptr);
   break;

  case DIPC_SEM_INFO:
   return get_sem_status(first, ptr);
   break;

  case DIPC_SHM_INFO:
   return get_shm_status(first, ptr);
   break; 

   case DIPC_MSG_INFO:
   return get_msg_status(first, ptr);
   break;

  default:
   printk("kernel: Error: unknown request\n");
   break;
 }   
 return DIPC_FAIL;
}
