#define THREE_LEVEL
/*
 * linux/ipc/shm.c
 * Copyright (C) 1992, 1993 Krishna Balasubramanian
 *         Many improvements/fixes by Bruno Haible.
 * Replaced `struct shm_desc' by `struct vm_area_struct', July 1994.
 *
 * DIPC parts Copyright (C) Kamran Karimi
 */

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/ipc.h>
#include <linux/shm.h>
#include <linux/stat.h>
#include <linux/malloc.h>

#include <asm/segment.h>
#include <asm/pgtable.h>
/* DIPC s */
#include <linux/dipc.h>
/* DIPC e */

extern int ipcperms (struct ipc_perm *ipcp, short shmflg);
extern unsigned int get_swap_page (void);
static int findkey (key_t key);
static int newseg (key_t key, int shmflg, int size);
static int shm_map (struct vm_area_struct *shmd);
static void killseg (int id);
static void shm_open (struct vm_area_struct *shmd);
static void shm_close (struct vm_area_struct *shmd);
static pte_t shm_swap_in(struct vm_area_struct *, unsigned long, unsigned long);

/* DIPC s */
int dipc_shmctl(int shmid, int cmd, struct shmid_ds *buff);
/* DIPC e */

static int shm_tot = 0; /* total number of shared memory pages */
static int shm_rss = 0; /* number of shared memory pages that are in memory */
static int shm_swp = 0; /* number of shared memory pages that are in swap */
static int max_shmid = 0; /* every used id is <= max_shmid */
static struct wait_queue *shm_lock = NULL; /* calling findkey() may need to wait */
static struct shmid_ds *shm_segs[SHMMNI];

static unsigned short shm_seq = 0; /* incremented, for recognizing stale ids */

/* some statistics */
static ulong swap_attempts = 0;
static ulong swap_successes = 0;
static ulong used_segs = 0;

void shm_init (void)
{
	int id;

	for (id = 0; id < SHMMNI; id++)
		shm_segs[id] = (struct shmid_ds *) IPC_UNUSED;		
	shm_tot = shm_rss = shm_seq = max_shmid = used_segs = 0;
	shm_lock = NULL;
	return;
}

static int findkey (key_t key)
{
	int id;
	struct shmid_ds *shp;

	for (id = 0; id <= max_shmid; id++) {
		while ((shp = shm_segs[id]) == IPC_NOID)
			sleep_on (&shm_lock);
		if (shp == IPC_UNUSED)
			continue;
		if (key == shp->shm_perm.key)
			return id;
	}
	return -1;
}



/*
 * allocate new shmid_ds and pgtable. protected by shm_segs[id] = NOID.
 */
static int newseg (key_t key, int shmflg, int size)
{
	struct shmid_ds *shp;
	int numpages = (size + PAGE_SIZE -1) >> PAGE_SHIFT;
	int id, i;

	if (size < SHMMIN)
		return -EINVAL;
	if (shm_tot + numpages >= SHMALL)
		return -ENOSPC;
	for (id = 0; id < SHMMNI; id++)
		if (shm_segs[id] == IPC_UNUSED) {
			shm_segs[id] = (struct shmid_ds *) IPC_NOID;
			goto found;
		}
	return -ENOSPC;

found:
	shp = (struct shmid_ds *) kmalloc (sizeof (*shp), GFP_KERNEL);
	if (!shp) {
		shm_segs[id] = (struct shmid_ds *) IPC_UNUSED;
		if (shm_lock)
			wake_up (&shm_lock);
		return -ENOMEM;
	}

	shp->shm_pages = (ulong *) kmalloc (numpages*sizeof(ulong),GFP_KERNEL);
	if (!shp->shm_pages) {
		shm_segs[id] = (struct shmid_ds *) IPC_UNUSED;
		if (shm_lock)
			wake_up (&shm_lock);
		kfree(shp);
		return -ENOMEM;
	}

	for (i = 0; i < numpages; shp->shm_pages[i++] = 0);
	shm_tot += numpages;
	shp->shm_perm.key = key;
	
	shp->shm_perm.cuid = shp->shm_perm.uid = current->euid;
	shp->shm_perm.cgid = shp->shm_perm.gid = current->egid;
	shp->shm_perm.seq = shm_seq;
/* DIPC s */	
	shp->shm_perm.mode = (shmflg & (S_IRWXUGO | IPC_DIPC));
/* DIPC e */	
	shp->shm_segsz = size;
	shp->shm_cpid = current->pid;
	shp->attaches = NULL;
	shp->shm_lpid = shp->shm_nattch = 0;
	shp->shm_atime = shp->shm_dtime = 0;
	shp->shm_ctime = CURRENT_TIME;
	shp->shm_npages = numpages;

	if (id > max_shmid)
		max_shmid = id;
	shm_segs[id] = shp;
	used_segs++;
	if (shm_lock)
		wake_up (&shm_lock);
	return (unsigned int) shp->shm_perm.seq * SHMMNI + id;
}


int sys_shmget (key_t key, int size, int shmflg)
{
	struct shmid_ds *shp;
	int id = 0;
	
	if (size < 0 || size > SHMMAX)
		return -EINVAL;
	if (key == IPC_PRIVATE)
		return newseg(key, shmflg, size);

/* DIPC s */
	/* The key may be locked and freed several times.
	   Also, another worker may need to access this SHM
	   while locked (updating its contents) */
	if(dipc_present() && !dipc_manager(current)) 
	 while(dipc_check(key, SHM));	
/* DIPC e */	 
	if ((id = findkey (key)) == -1) {
/* DIPC s */
		if(dipc_present() && !dipc_manager(current))
		{
		 int remote_size, result, distrib = shmflg & IPC_DIPC;
		 
		 dipc_excl(key, SHM);
		 result = dipc_search_key(key, SHM, distrib, &remote_size);
		 /**printk("Key %d has real_size %d\n", key, remote_size);*/
		 if(remote_size)
		 {
		  if( ((shmflg & IPC_DIPC) && (result == FOUND_IPC)) ||
		      (!(shmflg & IPC_DIPC) && (result == FOUND_DIPC)))
		  {
		   dipc_commit_key(key, SHM, 0, 0);
		   dipc_wake(key, SHM);    
		   return -EEXIST;
		  }
		  else if((shmflg & IPC_DIPC) && (result == FOUND_DIPC))
		  {  /* local and remote are DIPC, test remotely */
		   result = dipc_test_key(key, SHM, size, shmflg);
		   if(result < 0)
		   {
		    dipc_commit_key(key, SHM, 0, 0);
		    dipc_wake(key, SHM);
		    return result;
		   }
		  }
		  /* else local and remote are ipc, treat normally 
		  {
		   if (!(shmflg & IPC_CREAT))
		   return -ENOENT;
		   return newseg(key, shmflg, size);
		  } */ 
		 }  
		 else
		 { /* nothing was found in other machines */
		  if (!(shmflg & IPC_CREAT))
		  {
		   dipc_commit_key(key, SHM, 0, 0);
		   dipc_wake(key, SHM);
		   return -ENOENT;
		  }	
		 }
		 /* take the size as specified remotely */
		 if(remote_size)
		  size = remote_size;
		 id = newseg(key, shmflg, size);
		 if(id < 0) 
		 {
		  /**printk("Kernel: failed to create! size = %d\n",size);*/
		  dipc_commit_key(key, SHM, 0, 0);
		 } 
		 else
		 {
		  shp = shm_segs[id % SHMMNI];
		  if(shmflg & IPC_DIPC)
		  {
		   if(remote_size) /* this is not the owner */
		    shp->shm_perm.mode |= (IPC_ABS | IPC_WRP);
		   else
		    shp->shm_perm.mode |= IPC_OWN; 
		  } 
		  dipc_commit_key(key, SHM, size, shp->shm_perm.mode);  
		 }
		 dipc_wake(key, SHM);
		 return id;	
		}
		else /* no manager */
		{	 		
		 if (!(shmflg & IPC_CREAT))
			return -ENOENT;
		 return newseg(key, shmflg, size);
		} 
/* DIPC e */		
	}	
	if ((shmflg & IPC_CREAT) && (shmflg & IPC_EXCL))
		return -EEXIST;
	shp = shm_segs[id];
	
	if (size > shp->shm_segsz)
		return -EINVAL;
	if (ipcperms (&shp->shm_perm, shmflg))
		return -EACCES;
/* DIPC s */		
	if(((shp->shm_perm.mode & IPC_DIPC) && !(shmflg & IPC_DIPC)) ||
	   (!(shp->shm_perm.mode & IPC_DIPC) && (shmflg & IPC_DIPC)))
	 return -EEXIST;
/* DIPC e */	
	if (shp->shm_perm.mode & SHM_DEST)
/* DIPC s */
	{
	 /* this is to let access to SHM when it is removed but not
	    destroyed */
	 if(dipc_present() && dipc_manager(current))
	  return (unsigned int) shp->shm_perm.seq * SHMMNI + id;  	
	 else 
	  return -EIDRM; 
	}
/* DIPC e */		  		
	/*	return -EIDRM; */	
	return (unsigned int) shp->shm_perm.seq * SHMMNI + id;
}

/*
 * Only called after testing nattch and SHM_DEST.
 * Here pages, pgtable and shmid_ds are freed.
 */
static void killseg (int id)
{
	struct shmid_ds *shp;
	int i, numpages;
	
	shp = shm_segs[id];
	if (shp == IPC_NOID || shp == IPC_UNUSED) {
		printk ("shm nono: killseg called on unused seg id=%d\n", id);
		return;
	}
/* DIPC s */
	if(dipc_present())
	 while(dipc_check(shp->shm_perm.key, SHM));
/* DIPC e */		
	shp->shm_perm.seq++;     /* for shmat */
	shm_seq = (shm_seq+1) % ((unsigned)(1<<31)/SHMMNI); /* increment, but avoid overflow */
	shm_segs[id] = (struct shmid_ds *) IPC_UNUSED;
	used_segs--;
	if (id == max_shmid)
		while (max_shmid && (shm_segs[--max_shmid] == IPC_UNUSED));
	if (!shp->shm_pages) {
		printk ("shm nono: killseg shp->pages=NULL. id=%d\n", id);
		return;
	}
/* DIPC s */
	if(dipc_present())
	{
	 dipc_excl(shp->shm_perm.key, SHM);
	 /* if(shp->shm_perm.mode & IPC_OWN)*/
	 dipc_del_key(shp->shm_perm.key, SHM);	
	 dipc_wake(shp->shm_perm.key, SHM);
	} 
/* DIPC e */
	numpages = shp->shm_npages;
	for (i = 0; i < numpages ; i++) {
		pte_t pte;
		pte_val(pte) = shp->shm_pages[i];
		if (pte_none(pte))
			continue;
		if (pte_present(pte)) {
			free_page (pte_page(pte));
			shm_rss--;
		} else {
			swap_free(pte_val(pte));
			shm_swp--;
		}
	}
	kfree(shp->shm_pages);
	shm_tot -= numpages;
	kfree(shp);
	return;
}


int sys_shmctl (int shmid, int cmd, struct shmid_ds *buf)
{
	struct shmid_ds tbuf;
	struct shmid_ds *shp;
	struct ipc_perm *ipcp;
	int id, err;

	if (cmd < 0 || shmid < 0)
		return -EINVAL;
/* DIPC s */
	if(dipc_present() && !dipc_manager(current))
	{
	 if(cmd == IPC_STAT || cmd == IPC_SET || cmd == IPC_RMID)
	 {
	  shp = shm_segs[id = (unsigned int) shmid % SHMMNI];
	  if (shp == IPC_UNUSED || shp == IPC_NOID)
	 	return -EINVAL;
	  if (shp->shm_perm.seq != (unsigned int) shmid / SHMMNI)
		return -EIDRM;	
	  if((shp->shm_perm.mode & IPC_DIPC) && !(shp->shm_perm.mode & IPC_OWN))
	  {
	   return dipc_shmctl(id, cmd, buf);	
	  }
	 }
	}      
/* DIPC e */
 
	if (cmd == IPC_SET) {
		if (!buf)
			return -EFAULT;
		err = verify_area (VERIFY_READ, buf, sizeof (*buf));
		if (err)
			return err;
		memcpy_fromfs (&tbuf, buf, sizeof (*buf));
	}

	switch (cmd) { /* replace with proc interface ? */
	case IPC_INFO:
	{
		struct shminfo shminfo;
		if (!buf)
			return -EFAULT;
		shminfo.shmmni = SHMMNI;
		shminfo.shmmax = SHMMAX;
		shminfo.shmmin = SHMMIN;
		shminfo.shmall = SHMALL;
		shminfo.shmseg = SHMSEG;
		err = verify_area (VERIFY_WRITE, buf, sizeof (struct shminfo));
		if (err)
			return err;
		memcpy_tofs (buf, &shminfo, sizeof(struct shminfo));
		return max_shmid;
	}
	case SHM_INFO:
	{
		struct shm_info shm_info;
		if (!buf)
			return -EFAULT;
		err = verify_area (VERIFY_WRITE, buf, sizeof (shm_info));
		if (err)
			return err;
		shm_info.used_ids = used_segs;
		shm_info.shm_rss = shm_rss;
		shm_info.shm_tot = shm_tot;
		shm_info.shm_swp = shm_swp;
		shm_info.swap_attempts = swap_attempts;
		shm_info.swap_successes = swap_successes;
		memcpy_tofs (buf, &shm_info, sizeof(shm_info));
		return max_shmid;
	}
	case SHM_STAT:
		if (!buf)
			return -EFAULT;
		err = verify_area (VERIFY_WRITE, buf, sizeof (*buf));
		if (err)
			return err;
		if (shmid > max_shmid)
			return -EINVAL;
		shp = shm_segs[shmid];
		if (shp == IPC_UNUSED || shp == IPC_NOID)
			return -EINVAL;
		if (ipcperms (&shp->shm_perm, S_IRUGO))
			return -EACCES;
		id = (unsigned int) shp->shm_perm.seq * SHMMNI + shmid;
		tbuf.shm_perm   = shp->shm_perm;
		tbuf.shm_segsz  = shp->shm_segsz;
		tbuf.shm_atime  = shp->shm_atime;
		tbuf.shm_dtime  = shp->shm_dtime;
		tbuf.shm_ctime  = shp->shm_ctime;
		tbuf.shm_cpid   = shp->shm_cpid;
		tbuf.shm_lpid   = shp->shm_lpid;
		tbuf.shm_nattch = shp->shm_nattch;
		memcpy_tofs (buf, &tbuf, sizeof(*buf));
		return id;
	}
/* DIPC s */
	again:
/* DIPC e */
	shp = shm_segs[id = (unsigned int) shmid % SHMMNI];
	if (shp == IPC_UNUSED || shp == IPC_NOID)
		return -EINVAL;
	if (shp->shm_perm.seq != (unsigned int) shmid / SHMMNI)
		return -EIDRM;
	ipcp = &shp->shm_perm;
/* DIPC s */
	if(dipc_present())
	 if(dipc_check(shp->shm_perm.key, SHM)) goto again;
/* DIPC e */
	switch (cmd) {
	case SHM_UNLOCK:
		if (!suser())
			return -EPERM;
		if (!(ipcp->mode & SHM_LOCKED))
			return -EINVAL;
		ipcp->mode &= ~SHM_LOCKED;
		break;
	case SHM_LOCK:
/* Allow superuser to lock segment in memory */
/* Should the pages be faulted in here or leave it to user? */
/* need to determine interaction with current->swappable */
		if (!suser())
			return -EPERM;
		if (ipcp->mode & SHM_LOCKED)
			return -EINVAL;
		ipcp->mode |= SHM_LOCKED;
		break;
	case IPC_STAT:
		if (ipcperms (ipcp, S_IRUGO))
			return -EACCES;
		if (!buf)
			return -EFAULT;
		err = verify_area (VERIFY_WRITE, buf, sizeof (*buf));
		if (err)
			return err;
		tbuf.shm_perm   = shp->shm_perm;
		tbuf.shm_segsz  = shp->shm_segsz;
		tbuf.shm_atime  = shp->shm_atime;
		tbuf.shm_dtime  = shp->shm_dtime;
		tbuf.shm_ctime  = shp->shm_ctime;
		tbuf.shm_cpid   = shp->shm_cpid;
		tbuf.shm_lpid   = shp->shm_lpid;
		tbuf.shm_nattch = shp->shm_nattch;
		memcpy_tofs (buf, &tbuf, sizeof(*buf));
		break;
	case IPC_SET:
		if (suser() || current->euid == shp->shm_perm.uid ||
		    current->euid == shp->shm_perm.cuid) {
			ipcp->uid = tbuf.shm_perm.uid;
			ipcp->gid = tbuf.shm_perm.gid;
			ipcp->mode = (ipcp->mode & ~S_IRWXUGO)
				| (tbuf.shm_perm.mode & S_IRWXUGO);
			shp->shm_ctime = CURRENT_TIME;
			break;
		}
		return -EPERM;
	case IPC_RMID:
		if (suser() || current->euid == shp->shm_perm.uid ||
		    current->euid == shp->shm_perm.cuid) {
/* DIPC s */
			if(dipc_present())
			{
			 dipc_excl(shp->shm_perm.key, SHM);
			
		         if((shp->shm_perm.mode & IPC_DIPC) &&
		            (shp->shm_perm.mode & IPC_OWN))
		          dipc_rmid_key(shp->shm_perm.key, SHM);	
		         dipc_wake(shp->shm_perm.key, SHM);
		        } 
/* DIPC e */		
			shp->shm_perm.mode |= SHM_DEST;	
			if (shp->shm_nattch <= 0)
				killseg (id);
			break;
		}
		return -EPERM;
	default:
		return -EINVAL;
	}
	return 0;
}

/*
 * The per process internal structure for managing segments is
 * `struct vm_area_struct'.
 * A shmat will add to and shmdt will remove from the list.
 * shmd->vm_task	the attacher
 * shmd->vm_start	virt addr of attach, multiple of SHMLBA
 * shmd->vm_end		multiple of SHMLBA
 * shmd->vm_next	next attach for task
 * shmd->vm_next_share	next attach for segment
 * shmd->vm_offset	offset into segment
 * shmd->vm_pte		signature for this attach
 */

static struct vm_operations_struct shm_vm_ops = {
	shm_open,		/* open */
	shm_close,		/* close */
	NULL,			/* unmap */
	NULL,			/* protect */
	NULL,			/* sync */
	NULL,			/* advise */
	NULL,			/* nopage (done with swapin) */
	NULL,			/* wppage */
	NULL,			/* swapout (hardcoded right now) */
	shm_swap_in		/* swapin */
};

/* Insert shmd into the circular list shp->attaches */
static inline void insert_attach (struct shmid_ds * shp, struct vm_area_struct * shmd)
{
	struct vm_area_struct * attaches;

	if ((attaches = shp->attaches)) {
		shmd->vm_next_share = attaches;
		shmd->vm_prev_share = attaches->vm_prev_share;
		shmd->vm_prev_share->vm_next_share = shmd;
		attaches->vm_prev_share = shmd;
	} else
		shp->attaches = shmd->vm_next_share = shmd->vm_prev_share = shmd;
}

/* Remove shmd from circular list shp->attaches */
static inline void remove_attach (struct shmid_ds * shp, struct vm_area_struct * shmd)
{
	if (shmd->vm_next_share == shmd) {
		if (shp->attaches != shmd) {
			printk("shm_close: shm segment (id=%ld) attach list inconsistent\n",
				(shmd->vm_pte >> SHM_ID_SHIFT) & SHM_ID_MASK);
			printk("shm_close: %d %08lx-%08lx %c%c%c%c %08lx %08lx\n",
				shmd->vm_task->pid, shmd->vm_start, shmd->vm_end,
				shmd->vm_flags & VM_READ ? 'r' : '-',
				shmd->vm_flags & VM_WRITE ? 'w' : '-',
				shmd->vm_flags & VM_EXEC ? 'x' : '-',
				shmd->vm_flags & VM_MAYSHARE ? 's' : 'p',
				shmd->vm_offset, shmd->vm_pte);
		}
		shp->attaches = NULL;
	} else {
		if (shp->attaches == shmd)
			shp->attaches = shmd->vm_next_share;
		shmd->vm_prev_share->vm_next_share = shmd->vm_next_share;
		shmd->vm_next_share->vm_prev_share = shmd->vm_prev_share;
	}
}

/*
 * ensure page tables exist
 * mark page table entries with shm_sgn.
 */
static int shm_map (struct vm_area_struct *shmd)
{
	pgd_t *page_dir;
	pmd_t *page_middle;
	pte_t *page_table;
	unsigned long tmp, shm_sgn;

	/* clear old mappings */
	do_munmap(shmd->vm_start, shmd->vm_end - shmd->vm_start);

	/* add new mapping */
	insert_vm_struct(current, shmd);
	merge_segments(current, shmd->vm_start, shmd->vm_end);

	/* map page range */
	shm_sgn = shmd->vm_pte + ((shmd->vm_offset >> PAGE_SHIFT) << SHM_IDX_SHIFT);
	for (tmp = shmd->vm_start; tmp < shmd->vm_end; tmp += PAGE_SIZE,
	     shm_sgn += (1 << SHM_IDX_SHIFT)) {
		page_dir = pgd_offset(shmd->vm_task,tmp);
		page_middle = pmd_alloc(page_dir,tmp);
		if (!page_middle)
			return -ENOMEM;
		page_table = pte_alloc(page_middle,tmp);
		if (!page_table)
			return -ENOMEM;
		pte_val(*page_table) = shm_sgn;
	}
	invalidate();
	return 0;
}

/*
 * Fix shmaddr, allocate descriptor, map shm, add attach descriptor to lists.
 */
int sys_shmat (int shmid, char *shmaddr, int shmflg, ulong *raddr)
{
	struct shmid_ds *shp;
	struct vm_area_struct *shmd;
	int err;
	unsigned int id;
	unsigned long addr;

	if (shmid < 0) {
		/* printk("shmat() -> EINVAL because shmid = %d < 0\n",shmid); */
		return -EINVAL;
	}

	shp = shm_segs[id = (unsigned int) shmid % SHMMNI];
	if (shp == IPC_UNUSED || shp == IPC_NOID) {
		/* printk("shmat() -> EINVAL because shmid = %d is invalid\n",shmid); */
		return -EINVAL;
	}

	if (!(addr = (ulong) shmaddr)) {
		if (shmflg & SHM_REMAP)
			return -EINVAL;
		if (!(addr = get_unmapped_area(shp->shm_segsz)))
			return -ENOMEM;
	} else if (addr & (SHMLBA-1)) {
		if (shmflg & SHM_RND)
			addr &= ~(SHMLBA-1);       /* round down */
		else
			return -EINVAL;
	}
	if ((addr > current->mm->start_stack - 16384 - PAGE_SIZE*shp->shm_npages)) {
		/* printk("shmat() -> EINVAL because segment intersects stack\n"); */
		return -EINVAL;
	}
	if (!(shmflg & SHM_REMAP))
		if ((shmd = find_vma_intersection(current, addr, addr + shp->shm_segsz))) {
			/* printk("shmat() -> EINVAL because the interval [0x%lx,0x%lx) intersects an already mapped interval [0x%lx,0x%lx).\n",
				addr, addr + shp->shm_segsz, shmd->vm_start, shmd->vm_end); */
			return -EINVAL;
		}

	if (ipcperms(&shp->shm_perm, shmflg & SHM_RDONLY ? S_IRUGO : S_IRUGO|S_IWUGO))
		return -EACCES;
	if (shp->shm_perm.seq != (unsigned int) shmid / SHMMNI)
		return -EIDRM;

	shmd = (struct vm_area_struct *) kmalloc (sizeof(*shmd), GFP_KERNEL);
	if (!shmd)
		return -ENOMEM;
	if ((shp != shm_segs[id]) || (shp->shm_perm.seq != (unsigned int) shmid / SHMMNI)) {
		kfree(shmd);
		return -EIDRM;
	}

	shmd->vm_pte = (SHM_SWP_TYPE << 1) | (id << SHM_ID_SHIFT);
	shmd->vm_start = addr;
	shmd->vm_end = addr + shp->shm_npages * PAGE_SIZE;
	shmd->vm_task = current;
	shmd->vm_page_prot = (shmflg & SHM_RDONLY) ? PAGE_READONLY : PAGE_SHARED;

	shmd->vm_flags = VM_SHM | VM_MAYSHARE | VM_SHARED
			 | VM_MAYREAD | VM_MAYEXEC | VM_READ | VM_EXEC
			 | ((shmflg & SHM_RDONLY) ? 0 : VM_MAYWRITE | VM_WRITE);
/* DIPC s */	
	if(dipc_present() && !dipc_manager(shmd->vm_task))
	{
	 if(shp->shm_perm.mode & IPC_WRP)
	  shmd->vm_page_prot = 
	   __pgprot(pgprot_val(shmd->vm_page_prot) & ~_PAGE_RW);
	}   
/* DIPC e */	
	shmd->vm_next_share = shmd->vm_prev_share = NULL;
	shmd->vm_inode = NULL;
	shmd->vm_offset = 0;
	shmd->vm_ops = &shm_vm_ops;

	shp->shm_nattch++;            /* prevent destruction */
	if ((err = shm_map (shmd))) {
		if (--shp->shm_nattch <= 0 && shp->shm_perm.mode & SHM_DEST)
			killseg(id);
		kfree(shmd);
		return err;
	}

	insert_attach(shp,shmd);  /* insert shmd into shp->attaches */

	shp->shm_lpid = current->pid;
	shp->shm_atime = CURRENT_TIME;

	*raddr = addr;
	return 0;
}

/* This is called by fork, once for every shm attach. */
static void shm_open (struct vm_area_struct *shmd)
{
	unsigned int id;
	struct shmid_ds *shp;

	id = (shmd->vm_pte >> SHM_ID_SHIFT) & SHM_ID_MASK;
	shp = shm_segs[id];
	if (shp == IPC_UNUSED) {
		printk("shm_open: unused id=%d PANIC\n", id);
		return;
	}
	insert_attach(shp,shmd);  /* insert shmd into shp->attaches */
	shp->shm_nattch++;
	shp->shm_atime = CURRENT_TIME;
	shp->shm_lpid = current->pid;
}

/*
 * remove the attach descriptor shmd.
 * free memory for segment if it is marked destroyed.
 * The descriptor has already been removed from the current->mm->mmap list
 * and will later be kfree()d.
 */
static void shm_close (struct vm_area_struct *shmd)
{
	struct shmid_ds *shp;
	int id;

	unmap_page_range (shmd->vm_start, shmd->vm_end - shmd->vm_start);

	/* remove from the list of attaches of the shm segment */
	id = (shmd->vm_pte >> SHM_ID_SHIFT) & SHM_ID_MASK;
	shp = shm_segs[id];
	remove_attach(shp,shmd);  /* remove from shp->attaches */
  	shp->shm_lpid = current->pid;
	shp->shm_dtime = CURRENT_TIME;
	if (--shp->shm_nattch <= 0 && shp->shm_perm.mode & SHM_DEST)
		 killseg (id);
}

/*
 * detach and kill segment if marked destroyed.
 * The work is done in shm_close.
 */
int sys_shmdt (char *shmaddr)
{
	struct vm_area_struct *shmd, *shmdnext;

	for (shmd = current->mm->mmap; shmd; shmd = shmdnext) {
		shmdnext = shmd->vm_next;
		if (shmd->vm_ops == &shm_vm_ops
		    && shmd->vm_start - shmd->vm_offset == (ulong) shmaddr)
			do_munmap(shmd->vm_start, shmd->vm_end - shmd->vm_start);
	}
	return 0;
}

/*
 * page not present ... go through shm_pages
 */
static pte_t shm_swap_in(struct vm_area_struct * shmd, unsigned long offset, unsigned long code)
{
	pte_t pte;
	struct shmid_ds *shp;
	unsigned int id, idx;

	id = (code >> SHM_ID_SHIFT) & SHM_ID_MASK;
	if (id != ((shmd->vm_pte >> SHM_ID_SHIFT) & SHM_ID_MASK)) {
		printk ("shm_swap_in: code id = %d and shmd id = %ld differ in current %d\n",
			id, (shmd->vm_pte >> SHM_ID_SHIFT) & SHM_ID_MASK, current->pid);
		return BAD_PAGE;
	}
	if (id > max_shmid) {
		printk ("shm_swap_in: id=%d too big. proc mem corrupted\n", id);
		return BAD_PAGE;
	}
	shp = shm_segs[id];
	if (shp == IPC_UNUSED || shp == IPC_NOID) {
		printk ("shm_swap_in: id=%d invalid. Race.\n", id);
		return BAD_PAGE;
	}
	idx = (code >> SHM_IDX_SHIFT) & SHM_IDX_MASK;
	if (idx != (offset >> PAGE_SHIFT)) {
		printk ("shm_swap_in: code idx = %u and shmd idx = %lu differ\n",
			idx, offset >> PAGE_SHIFT);
		return BAD_PAGE;
	}
	if (idx >= shp->shm_npages) {
		printk ("shm_swap_in : too large page index. id=%d\n", id);
		return BAD_PAGE;
	}

	pte_val(pte) = shp->shm_pages[idx];
	if (!pte_present(pte)) {
		unsigned long page = get_free_page(GFP_KERNEL);
		if (!page) {
			oom(current);
			return BAD_PAGE;
		}
		pte_val(pte) = shp->shm_pages[idx];
		if (pte_present(pte)) {
			free_page (page); /* doesn't sleep */
			goto done;
		}
		if (!pte_none(pte)) {
			read_swap_page(pte_val(pte), (char *) page);
			pte_val(pte) = shp->shm_pages[idx];
			if (pte_present(pte))  {
				free_page (page); /* doesn't sleep */
				goto done;
			}
			swap_free(pte_val(pte));
			shm_swp--;
		}
		shm_rss++;
		pte = pte_mkdirty(mk_pte(page, PAGE_SHARED));
		shp->shm_pages[idx] = pte_val(pte);
	} else
		--current->mm->maj_flt;  /* was incremented in do_no_page */

done:	/* pte_val(pte) == shp->shm_pages[idx] */
	current->mm->min_flt++;
	mem_map[MAP_NR(pte_page(pte))]++;
	return pte_modify(pte, shmd->vm_page_prot);
}

/*
 * Goes through counter = (shm_rss >> prio) present shm pages.
 */
static unsigned long swap_id = 0; /* currently being swapped */
static unsigned long swap_idx = 0; /* next to swap */

int shm_swap (int prio)
{
	pte_t page;
	struct shmid_ds *shp;
	struct vm_area_struct *shmd;
	unsigned int swap_nr;
	unsigned long id, idx;
	int loop = 0, invalid = 0;
	int counter;

	counter = shm_rss >> prio;
	if (!counter || !(swap_nr = get_swap_page()))
		return 0;

 check_id:
	shp = shm_segs[swap_id];
	if (shp == IPC_UNUSED || shp == IPC_NOID || shp->shm_perm.mode & SHM_LOCKED ) {
		next_id:
		swap_idx = 0;
		if (++swap_id > max_shmid) {
			if (loop)
				goto failed;
			loop = 1;
			swap_id = 0;
		}
		goto check_id;
	}
	id = swap_id;

 check_table:
	idx = swap_idx++;
	if (idx >= shp->shm_npages)
		goto next_id;

	pte_val(page) = shp->shm_pages[idx];
	if (!pte_present(page))
		goto check_table;
	swap_attempts++;

	if (--counter < 0) { /* failed */
		failed:
		if (invalid)
			invalidate();
		swap_free (swap_nr);
		return 0;
	}
	if (shp->attaches)
	  for (shmd = shp->attaches; ; ) {
	    do {
		pgd_t *page_dir;
		pmd_t *page_middle;
		pte_t *page_table, pte;
		unsigned long tmp;

		if ((shmd->vm_pte >> SHM_ID_SHIFT & SHM_ID_MASK) != id) {
			printk ("shm_swap: id=%ld does not match shmd->vm_pte.id=%ld\n", id, shmd->vm_pte >> SHM_ID_SHIFT & SHM_ID_MASK);
			continue;
		}
		tmp = shmd->vm_start + (idx << PAGE_SHIFT) - shmd->vm_offset;
		if (!(tmp >= shmd->vm_start && tmp < shmd->vm_end))
			continue;
		page_dir = pgd_offset(shmd->vm_task,tmp);
		if (pgd_none(*page_dir) || pgd_bad(*page_dir)) {
			printk("shm_swap: bad pgtbl! id=%ld start=%lx idx=%ld\n",
					id, shmd->vm_start, idx);
			pgd_clear(page_dir);
			continue;
		}
		page_middle = pmd_offset(page_dir,tmp);
		if (pmd_none(*page_middle) || pmd_bad(*page_middle)) {
			printk("shm_swap: bad pgmid! id=%ld start=%lx idx=%ld\n",
					id, shmd->vm_start, idx);
			pmd_clear(page_middle);
			continue;
		}
		page_table = pte_offset(page_middle,tmp);
		pte = *page_table;
		if (!pte_present(pte))
			continue;
		if (pte_young(pte)) {
			*page_table = pte_mkold(pte);
			continue;
		}
		if (pte_page(pte) != pte_page(page))
			printk("shm_swap_out: page and pte mismatch\n");
		pte_val(*page_table) = shmd->vm_pte | idx << SHM_IDX_SHIFT;
		mem_map[MAP_NR(pte_page(pte))]--;
		if (shmd->vm_task->mm->rss > 0)
			shmd->vm_task->mm->rss--;
		invalid++;
	    /* continue looping through circular list */
	    } while (0);
	    if ((shmd = shmd->vm_next_share) == shp->attaches)
		break;
	}

	if (mem_map[MAP_NR(pte_page(page))] != 1)
		goto check_table;
	shp->shm_pages[idx] = swap_nr;
	if (invalid)
		invalidate();
	write_swap_page (swap_nr, (char *) pte_page(page));
	free_page(pte_page(page));
	swap_successes++;
	shm_swp++;
	shm_rss--;
	return 1;
}

/********************************************************************
 * Implement DIPC!
 */



/* this routine swaps out an entier shared memory. This constructed by
   doing some cut and paste on shm_swap. I hope this is done the right way! */ 
int dipc_swap_out (int id)
{
 pte_t page;
 struct shmid_ds *shp;
 struct vm_area_struct *shmd;
 unsigned int swap_nr;
 unsigned long idx;
 int invalid = 0;

 shp = shm_segs[id];
 if(shp == IPC_UNUSED || shp == IPC_NOID || shp->shm_perm.mode & SHM_LOCKED)
  return DIPC_FAIL;

 /* swap_free (swap_nr); */
		
 if (shp->attaches)
 {
  for(idx = 0; idx < shp->shm_npages; idx++)
  { 
   swap_nr = get_swap_page();
   if(swap_nr == 0) 
    return DIPC_FAIL;
   pte_val(page) = shp->shm_pages[idx];
   if(!pte_present(page))
    continue;
   swap_attempts++; 
   for (shmd = shp->attaches; ; )
   {
    do
    {
     pgd_t *page_dir;
     pmd_t *page_middle;
     pte_t *page_table, pte;
     unsigned long tmp;

     if ((shmd->vm_pte >> SHM_ID_SHIFT & SHM_ID_MASK) != id) 
     {
      printk ("dipc_swap_out: id=%d does not match shmd->vm_pte.id=%ld\n", 
              id, shmd->vm_pte >> SHM_ID_SHIFT & SHM_ID_MASK);
      continue;
     }
     tmp = shmd->vm_start + (idx << PAGE_SHIFT) - shmd->vm_offset;
     if (!(tmp >= shmd->vm_start && tmp < shmd->vm_end))
      continue;
     page_dir = pgd_offset(shmd->vm_task,tmp);
     if (pgd_none(*page_dir) || pgd_bad(*page_dir))
     {
      printk("dipc_swap_out: bad pgtbl! id=%d start=%lx idx=%ld\n",
					id, shmd->vm_start, idx);
      pgd_clear(page_dir);
      continue;
     }
     page_middle = pmd_offset(page_dir,tmp);
     if (pmd_none(*page_middle) || pmd_bad(*page_middle))
     {
      printk("dipc_swap_out: bad pgmid! id=%d start=%lx idx=%ld\n",
				id, shmd->vm_start, idx);
      pmd_clear(page_middle);
      continue;
     }
     page_table = pte_offset(page_middle,tmp);
     pte = *page_table;
     if (!pte_present(pte))
      continue;
     /*
     if (pte_young(pte))
     {
      *page_table = pte_mkold(pte);
      continue;
     }
     */
     if (pte_page(pte) != pte_page(page))
      printk("dipc_swap_out: page and pte mismatch\n");
     pte_val(*page_table) = shmd->vm_pte | idx << SHM_IDX_SHIFT;
     mem_map[MAP_NR(pte_page(pte))]--;
     if (shmd->vm_task->mm->rss > 0)
      shmd->vm_task->mm->rss--;
     invalid++;
     /* continue looping through circular list */
    } while (0);
    if ((shmd = shmd->vm_next_share) == shp->attaches)
     break;
   } /* for shmd... */
   /*
   if (mem_map[MAP_NR(pte_page(page))] != 1)
    goto check_table;
   */	
   shp->shm_pages[idx] = swap_nr;
   write_swap_page (swap_nr, (char *) pte_page(page));
   free_page(pte_page(page));
   swap_successes++;
   shm_swp++;
   shm_rss--;
  } /* for idx ... */	
  invalidate();
 } /* if shmat... */
 return DIPC_SUCCESS;
}


/* the following three routines are copied from mm/mprotect.c.
   They were declared as static and I did not want to change it */
static inline void change_pte_range(pmd_t * pmd, unsigned long address,
	unsigned long size, pgprot_t newprot)
{
	pte_t * pte;
	unsigned long end;

	if (pmd_none(*pmd))
		return;
	if (pmd_bad(*pmd)) {
		printk("dipcd: change_pte_range: bad pmd (%08lx)\n", pmd_val(*pmd));
		pmd_clear(pmd);
		return;
	}
	pte = pte_offset(pmd, address);
	address &= ~PMD_MASK;
	end = address + size;
	if (end > PMD_SIZE)
		end = PMD_SIZE;
	do {
		pte_t entry = *pte;
		if (pte_present(entry))
			*pte = pte_modify(entry, newprot);
		address += PAGE_SIZE;
		pte++;
	} while (address < end);
}

static inline void change_pmd_range(pgd_t * pgd, unsigned long address,
	unsigned long size, pgprot_t newprot)
{
	pmd_t * pmd;
	unsigned long end;

	if (pgd_none(*pgd))
		return;
	if (pgd_bad(*pgd)) {
		printk("dipcd: change_pmd_range: bad pgd (%08lx)\n", pgd_val(*pgd));
		pgd_clear(pgd);
		return;
	}
	pmd = pmd_offset(pgd, address);
	address &= ~PGDIR_MASK;
	end = address + size;
	if (end > PGDIR_SIZE)
		end = PGDIR_SIZE;
	do {
		change_pte_range(pmd, address, end - address, newprot);
		address = (address + PMD_SIZE) & PMD_MASK;
		pmd++;
	} while (address < end);
}

static void change_protection(struct task_struct *tsk,
                           struct vm_area_struct *vma, pgprot_t newprot)
{
        unsigned long start,end;
	pgd_t *dir;

	start = vma->vm_start;
	end = vma->vm_end;
	dir = pgd_offset(tsk, start);
	while (start < end) {
		change_pmd_range(dir, start, end - start, newprot);
		start = (start + PGDIR_SIZE) & PGDIR_MASK;
		dir++;
	}
	vma->vm_page_prot = newprot;
/*
	if(pgprot_val(newprot) & _PAGE_RW)
	 vma->vm_flags |= (VM_MAYWRITE | VM_WRITE);
	else
	 vma->vm_flags &= ~(VM_MAYWRITE | VM_WRITE);
*/ 
	invalidate();
	return;
 }

static int enable_prot(key_t key, int enable_write)
{
 int id;
 struct shmid_ds *shp;
 struct vm_area_struct *attaches;
 
 id = findkey(key);
 if(id < 0) return -1;
 shp = shm_segs[id];
 
 if(enable_write)
  shp->shm_perm.mode &= ~(IPC_WRP | IPC_ABS);
 else
  shp->shm_perm.mode &= ~IPC_ABS;
 attaches = shp->attaches;
 while(attaches)
 { 
  if(enable_write)
  {
   /* only if the segment had this right in the first place */
   if(attaches->vm_flags & VM_WRITE)
   {
    /**printk("Kernel: setprot: protecting: task No. %d\n",attaches->vm_task->pid);*/
    change_protection(attaches->vm_task,attaches,
             __pgprot(pgprot_val(attaches->vm_page_prot) | _PAGE_RW));
   }
   if(dipc_present() && !dipc_manager(attaches->vm_task))
   {  
    /**printk("Kernel: WRITE SIG %d to %d\n",DIPC_SIG_WRITER,attaches->vm_task->pid);*/
    send_sig(DIPC_SIG_WRITER, attaches->vm_task, 1);                         
   } 
  }
  else if(dipc_present() && !dipc_manager(attaches->vm_task))   
  {
   /**printk("Kernel: READ SIG %d to %d\n",DIPC_SIG_READER,attaches->vm_task->pid);*/
   send_sig(DIPC_SIG_READER, attaches->vm_task, 1);   
  } 
  attaches = attaches->vm_next_share;
  if(shp->attaches == attaches) 
   break;
 }
 return 0;
}

int disable_prot(key_t key, int cmd, int *mask)
{
 struct shmid_ds *shp;
 int id;
 int op, err;
 struct vm_area_struct *attaches;
 
 id = findkey(key);
 if(id == -1) 
 {
  printk("Kernel: key not found\n");
  return -ENOENT;
 }
 shp = shm_segs[id];
 err = verify_area(VERIFY_READ, mask, sizeof(int));
 if(err) return err;
 memcpy_fromfs(&op, mask, sizeof(int));
 /* for future attaches */
 if(op == DISABLE_WR)
 { 
  shp->shm_perm.mode |= IPC_WRP;
  attaches = shp->attaches;
  while(attaches)
  { 
   /**printk("Kernel: absenting: task No. %d\n",attaches->vm_task->pid);*/
   change_protection(attaches->vm_task,attaches,
             __pgprot(pgprot_val(attaches->vm_page_prot) & ~_PAGE_RW));
   attaches = attaches->vm_next_share;
   if(shp->attaches == attaches) /* circular */
    break;
  }
 }  
 else /* read protect */
 {
  shp->shm_perm.mode |= (IPC_ABS | IPC_WRP);
  if(dipc_swap_out(id) != DIPC_SUCCESS)
  {
   printk("DIPC Kernel: Can't swap-out the shared memory with key %d\n",key);
   return DIPC_FAIL;
  }
 }
 return DIPC_SUCCESS;
}



int dipc_shm_fault(struct vm_area_struct *vma, unsigned long write_access)
{
 int id;
 struct shmid_ds *shp;
 struct dipc_request *dr;
 key_t key;
 
 if(!(vma->vm_flags & VM_WRITE) && write_access) return 0;
 if(dipc_present() && !dipc_manager(vma->vm_task))
 { 
  id = (vma->vm_pte >> SHM_ID_SHIFT) & SHM_ID_MASK;
  shp = shm_segs[id];
  key = shp->shm_perm.key;
  while(dipc_check(shp->shm_perm.key, SHM));
  if((shp->shm_perm.mode & IPC_ABS) ||
     ((shp->shm_perm.mode & IPC_WRP) && (write_access)))
  { 
   dipc_excl(key, SHM);
   /* printk("dsm_changed: id: %d, mode: %d\n",id, shp->shm_perm.mode); */
   dr = kmalloc(sizeof(struct dipc_request), GFP_KERNEL);
   if(dr == NULL) return -ENOMEM; /* set errno */
   dr->info.key = key;
   dr->info.type = SHM;
   dr->info.pid = current->pid;
   dr->info.uid = current->uid;
   if(write_access)
    dr->info.func = WANTWR;
   else
    dr->info.func = WANTRD;
   dr->info.arg1 = key;
   dr->info.arg2 = 0;
   dr->info.arg3 = 0;
   dr->info.arg4 = 0;
   dr->info.owner = -1;
   dr->working = 0;
   dr->next = dr->prev = NULL;
   dr->lock = NULL;
   dr->info.dsize = 0;
   dr->data = NULL;
   
   dipc_add_request(dr);
   
   if(dr->info.arg1 < 0)
    send_sig(SIGSEGV, current, 1);
   else
    enable_prot(key, write_access);  
   kfree(dr); 
   dipc_wake(key, SHM);
   /**printk("Kernel: shm_changed: request replied!\n");*/
   return 1;
  }
 } 
 return 0; 
}


  
int dipc_shmctl(int shmid, int cmd, struct shmid_ds *buff)
{
 struct dipc_request *dr;
 int err, result;
 
 /**printk("Kernel: dipc_shmctl: starting\n");*/
 err = 0;
 if(cmd == IPC_SET)
  err = verify_area(VERIFY_READ, buff, sizeof(struct shmid_ds));
 else if(cmd == IPC_STAT)
  err = verify_area(VERIFY_WRITE, buff, sizeof(struct shmid_ds));
 
 if(err) return err;
 dr = kmalloc(sizeof(struct dipc_request), GFP_KERNEL);
 if(dr == NULL) return -ENOMEM; /* set errno */
 dr->info.key = shm_segs[shmid]->shm_perm.key;
 dr->info.type = SHM;
 dr->info.pid = current->pid;
 dr->info.uid = current->uid;
 dr->info.func = SHMCTL;
 dr->info.arg1 = shm_segs[shmid]->shm_perm.key;
 dr->info.arg2 = cmd;
 dr->info.arg3 = 0;
 dr->info.arg4 = 0;
 dr->info.owner = -1; 
 dr->info.dsize = 0;
 dr->working = 0;
 dr->next = dr->prev = NULL;
 dr->data = NULL;
 dr->lock = NULL;
 if((cmd == IPC_SET) || (cmd == IPC_STAT))
 {
  dr->info.dsize = sizeof(struct shmid_ds);
  dr->data = kmalloc(dr->info.dsize, GFP_KERNEL);
  if(dr->data == NULL)
  {
   kfree(dr);
   return -ENOMEM;
  }
 }
 if(cmd == IPC_SET)
  memcpy_fromfs(dr->data, buff, dr->info.dsize);

 /* printk("Kernel: dipc_shmctl: adding request\n"); */
 dipc_add_request(dr); 
 /* printk("dipc_shmctl: result came\n"); */
 result = dr->info.arg1;
 if(result == 0)
 {
  if(cmd == IPC_STAT)
   memcpy_tofs(buff, dr->data, dr->info.dsize);
 }
 if((cmd == IPC_STAT) || (cmd == IPC_SET))
  kfree(dr->data);
 kfree(dr);
 /* printk("dipc_shmctl: done!\n");  */
 return result;
} /* shmctl */


int shm_dipc(key_t key, int cmd, int *mask)
{
 struct shmid_ds *shp;
 int id;
 int op, err;
 struct vm_area_struct *attaches;
 
 id = findkey(key);
 if(id == -1) 
 {
  /**printk("Kernel: key not found\n");*/
  return -ENOENT;
 }
 shp = shm_segs[id];
 err = verify_area(VERIFY_READ, mask, sizeof(int));
 if(err) return err;
 memcpy_fromfs(&op, mask, sizeof(int));
 /* for future attaches */
 if(op == DISABLE_WR)
 { 
  shp->shm_perm.mode |= IPC_WRP;
  attaches = shp->attaches;
  if(attaches)
  { 
   do
   {
    /**printk("Kernel: absenting: task No. %d\n",attaches->vm_task->pid);*/
    change_protection(attaches->vm_task,attaches,
             __pgprot(pgprot_val(attaches->vm_page_prot) & ~_PAGE_RW));
    attaches = attaches->vm_next_share;
   }while(shp->attaches != attaches); /* circular */
  }
 }  
 else /* read protect */
 {
  shp->shm_perm.mode |= (IPC_ABS | IPC_WRP);
  dipc_swap_out(id); 
 }
 return 0;
}

long silent_find_owner_address(key_t key, int type);
 
int get_shm_status(int key, struct shmid_ds *ptr)
{
 int err,id;
 struct shmid_ds shm;
 
 id = findkey(key);
 if(id < 0)
  return -1;
 if(shm_segs[id] == IPC_UNUSED || shm_segs[id] == IPC_NOID)
  return -1;  
 err = verify_area(VERIFY_WRITE, ptr, sizeof(struct shmid_ds));
 if(err) return err;
 shm = *shm_segs[id];
 shm.shm_perm.seq = silent_find_owner_address(key, SHM);
 memcpy_tofs(ptr, &shm, sizeof(struct shmid_ds));
 return id;
}
